from __future__ import annotations

from time import sleep

from textual.notifications import Notification, Notifications


def test_empty_to_start_with() -> None:
    """We should have no notifications if we've not raised any."""
    assert len(Notifications()) == 0


def test_many_notifications() -> None:
    """Adding lots of long-timeout notifications should result in them being in the list."""
    tester = Notifications()
    for _ in range(100):
        tester.add(Notification("test", timeout=60))
    assert len(tester) == 100


def test_timeout() -> None:
    """Notifications should timeout from the list."""
    tester = Notifications()
    for n in range(100):
        tester.add(Notification("test", timeout=(0.5 if bool(n % 2) else 60)))
    assert len(tester) == 100
    sleep(0.6)
    assert len(tester) == 50


def test_in() -> None:
    """It should be possible to test if a notification is in a collection."""
    tester = Notifications()
    within = Notification("within", timeout=120)
    outwith = Notification("outwith", timeout=120)
    tester.add(within)
    assert within in tester
    assert outwith not in tester


def test_remove_notification() -> None:
    """It should be possible to remove a notification."""
    tester = Notifications()
    first = Notification("first", timeout=120)
    second = Notification("second", timeout=120)
    third = Notification("third", timeout=120)
    tester.add(first)
    tester.add(second)
    tester.add(third)
    assert list(tester) == [first, second, third]
    del tester[second]
    assert list(tester) == [first, third]
    del tester[first]
    assert list(tester) == [third]
    del tester[third]
    assert list(tester) == []


def test_remove_notification_multiple_times() -> None:
    """It should be possible to remove the same notification more than once without an error."""
    tester = Notifications()
    alert = Notification("delete me")
    tester.add(alert)
    assert list(tester) == [alert]
    del tester[alert]
    assert list(tester) == []
    del tester[alert]
    assert list(tester) == []


def test_clear() -> None:
    """It should be possible to clear all notifications."""
    tester = Notifications()
    for _ in range(100):
        tester.add(Notification("test", timeout=120))
    assert len(tester) == 100
    tester.clear()
    assert len(tester) == 0



from __future__ import annotations

import pytest

from textual.app import App, ComposeResult
from textual.color import Color
from textual.containers import Vertical
from textual.css.parse import parse
from textual.css.tokenizer import EOFError, TokenError
from textual.widgets import Button, Label


class NestedApp(App):
    CSS = """
    Screen {
        & > #foo {
            background: red;
            #egg {
                background: green;
            }
            .paul {
                background: blue;
            }
            &.jessica {
                color: magenta;
            }
        }
    }    
    """

    def compose(self) -> ComposeResult:
        with Vertical(id="foo", classes="jessica"):
            yield Label("Hello", id="egg")
            yield Label("World", classes="paul")


async def test_nest_app():
    """Test nested CSS works as expected."""
    app = NestedApp()
    async with app.run_test():
        assert app.query_one("#foo").styles.background == Color.parse("red")
        assert app.query_one("#foo").styles.color == Color.parse("magenta")
        assert app.query_one("#egg").styles.background == Color.parse("green")
        assert app.query_one("#foo .paul").styles.background == Color.parse("blue")


class ListOfNestedSelectorsApp(App[None]):
    CSS = """
    Label {
        &.foo, &.bar {
            background: red;
        }
    }
    """

    def compose(self) -> ComposeResult:
        yield Label("one", classes="foo")
        yield Label("two", classes="bar")
        yield Label("three", classes="heh")


async def test_lists_of_selectors_in_nested_css() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3969."""
    app = ListOfNestedSelectorsApp()
    red = Color.parse("red")
    async with app.run_test():
        assert app.query_one(".foo").styles.background == red
        assert app.query_one(".bar").styles.background == red
        assert app.query_one(".heh").styles.background != red


class DeclarationAfterNestedApp(App[None]):
    CSS = """
    Screen {
        background: green;
        Label {
            background: red;
        }        
    }
    """

    def compose(self) -> ComposeResult:
        yield Label("one")


async def test_rule_declaration_after_nested() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3999."""
    app = DeclarationAfterNestedApp()
    async with app.run_test():
        assert app.screen.styles.background == Color.parse("green")
        assert app.query_one(Label).styles.background == Color.parse("red")


@pytest.mark.parametrize(
    ("css", "exception"),
    [
        ("Selector {", EOFError),
        ("Selector{ Foo {", EOFError),
        ("Selector{ Foo {}", EOFError),
        ("> {}", TokenError),
        ("&", TokenError),
        ("&&", TokenError),
        ("&.foo", TokenError),
        ("& .foo", TokenError),
        ("{", TokenError),
        ("*{", EOFError),
    ],
)
def test_parse_errors(css: str, exception: type[Exception]) -> None:
    """Check some CSS which should fail."""
    with pytest.raises(exception):
        list(parse("", css, ("foo", "")))


class PseudoClassesInNestedApp(App[None]):
    CSS = """
    Vertical {
        Button:light, Button:dark {
            background: red;
        }

        min-height: 3;  # inconsequential rule to add entropy.

        #two, *:focus {
            background: green !important;
        }

        height: auto;  # inconsequential rule to add entropy.

        Label {
            background: yellow;

            &:light, &:dark {
                background: red;
            }

            &:hover {
                background: green !important;
            }
        }
    }
    """

    AUTO_FOCUS = "Button"

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Button(id="one", classes="first_half")
            yield Button(id="two", classes="first_half")
        with Vertical():
            yield Label("Hello, world!", id="three", classes="first_half")
            yield Label("Hello, world!", id="four", classes="first_half")
        with Vertical():
            yield Button(id="five", classes="second_half")
            yield Button(id="six", classes="second_half")
            yield Label("Hello, world!", id="seven", classes="second_half")
            yield Label("Hello, world!", id="eight", classes="second_half")


async def test_pseudo_classes_work_in_nested_css() -> None:
    """Makes sure pseudo-classes are correctly understood in nested TCSS.

    Regression test for https://github.com/Textualize/textual/issues/4039.
    """

    app = PseudoClassesInNestedApp()
    green = Color.parse("green")
    red = Color.parse("red")
    async with app.run_test() as pilot:
        assert app.query_one("#one").styles.background == green
        assert app.query_one("#two").styles.background == green
        assert app.query_one("#five").styles.background == red
        assert app.query_one("#six").styles.background == red

        assert app.query_one("#three").styles.background == red
        assert app.query_one("#four").styles.background == red
        assert app.query_one("#seven").styles.background == red
        assert app.query_one("#eight").styles.background == red

        await pilot.hover("#eight")

        assert app.query_one("#three").styles.background == red
        assert app.query_one("#four").styles.background == red
        assert app.query_one("#seven").styles.background == red
        assert app.query_one("#eight").styles.background == green



"""Test basic functioning of some containers."""

from textual.app import App, ComposeResult
from textual.containers import (
    Center,
    Horizontal,
    HorizontalScroll,
    Middle,
    Vertical,
    VerticalScroll,
)
from textual.widgets import Label


async def test_horizontal_vs_horizontalscroll_scrolling():
    """Check the default scrollbar behaviours for `Horizontal` and `HorizontalScroll`."""

    class HorizontalsApp(App[None]):
        CSS = """
        Screen {
            layout: vertical;
        }
        """

        def compose(self) -> ComposeResult:
            with Horizontal():
                for _ in range(10):
                    yield Label("How is life going? " * 3 + " | ")
            with HorizontalScroll():
                for _ in range(10):
                    yield Label("How is life going? " * 3 + " | ")

    WIDTH = 80
    HEIGHT = 24
    app = HorizontalsApp()
    async with app.run_test(size=(WIDTH, HEIGHT)):
        horizontal = app.query_one(Horizontal)
        horizontal_scroll = app.query_one(HorizontalScroll)
        assert horizontal.size.height == horizontal_scroll.size.height
        assert horizontal.scrollbars_enabled == (False, False)
        assert horizontal_scroll.scrollbars_enabled == (False, True)


async def test_vertical_vs_verticalscroll_scrolling():
    """Check the default scrollbar behaviours for `Vertical` and `VerticalScroll`."""

    class VerticalsApp(App[None]):
        CSS = """
        Screen {
            layout: horizontal;
        }
        """

        def compose(self) -> ComposeResult:
            with Vertical():
                for _ in range(10):
                    yield Label("How is life going?\n" * 3 + "\n\n")
            with VerticalScroll():
                for _ in range(10):
                    yield Label("How is life going?\n" * 3 + "\n\n")

    WIDTH = 80
    HEIGHT = 24
    app = VerticalsApp()
    async with app.run_test(size=(WIDTH, HEIGHT)):
        vertical = app.query_one(Vertical)
        vertical_scroll = app.query_one(VerticalScroll)
        assert vertical.size.width == vertical_scroll.size.width
        assert vertical.scrollbars_enabled == (False, False)
        assert vertical_scroll.scrollbars_enabled == (True, False)


async def test_center_container():
    """Check the size of the container `Center`."""

    class CenterApp(App[None]):
        def compose(self) -> ComposeResult:
            with Center():
                yield Label("<>\n<>\n<>")

    app = CenterApp()
    async with app.run_test():
        center = app.query_one(Center)
        assert center.size.width == app.size.width
        assert center.size.height == 3


async def test_middle_container():
    """Check the size of the container `Middle`."""

    class MiddleApp(App[None]):
        def compose(self) -> ComposeResult:
            with Middle():
                yield Label("1234")

    app = MiddleApp()
    async with app.run_test():
        middle = app.query_one(Middle)
        assert middle.size.width == 4
        assert middle.size.height == app.size.height


async def test_scrollbar_zero_thickness():
    """Ensuring that scrollbars can be set to zero thickness."""

    class ScrollbarZero(App):
        CSS = """* {
            scrollbar-size: 0 0;
            scrollbar-size-vertical: 0;  /* just exercising the parser */
            scrollbar-size-horizontal: 0;  /* exercise the parser */
        }
        """

        def compose(self) -> ComposeResult:
            with Vertical():
                for _ in range(10):
                    yield Label("Hello, world!")

    app = ScrollbarZero()
    async with app.run_test(size=(8, 6)):
        pass



from dataclasses import dataclass
from functools import partial

from textual import on
from textual._color_constants import COLOR_NAME_TO_RGB
from textual.app import App, ComposeResult
from textual.command import Hit, Hits, Provider
from textual.message import Message
from textual.widgets import Header, Static


@dataclass
class SwitchColor(Message, bubble=False):
    """Message to tell the app to switch color."""

    color: str


class ColorCommands(Provider):
    """A command provider to select colors."""

    async def search(self, query: str) -> Hits:
        """Called for each key."""
        matcher = self.matcher(query)
        for color in COLOR_NAME_TO_RGB.keys():
            score = matcher.match(color)
            if score > 0:
                yield Hit(
                    score,
                    matcher.highlight(color),
                    partial(self.app.post_message, SwitchColor(color)),
                )


class ColorBlock(Static):
    """Simple block of color."""

    DEFAULT_CSS = """
    ColorBlock{
        padding: 3 6;
        margin: 1 2;
        color: auto;
    }
    """


class ColorApp(App):
    """Experiment with the command palette."""

    COMMANDS = App.COMMANDS | {ColorCommands}
    TITLE = "Press ctrl + p and type a color"

    def compose(self) -> ComposeResult:
        yield Header()

    @on(SwitchColor)
    def switch_color(self, event: SwitchColor) -> None:
        """Adds a color block on demand."""
        color_block = ColorBlock(event.color)
        color_block.styles.background = event.color
        self.mount(color_block)
        self.screen.scroll_end()


if __name__ == "__main__":
    app = ColorApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import HorizontalScroll, VerticalScroll
from textual.widgets import Label


class MyApp(App[None]):
    AUTO_FOCUS = ""
    CSS = """
    VerticalScroll {
        border: round $primary;
    }
    #vertical {
        height: 21;
    }
    HorizontalScroll {
        border: round $accent;
        height: auto;
    }
    Label {
        height: auto;
        width: auto;
    }
    """

    def compose(self) -> ComposeResult:
        yield Label(("SPAM\n" * 53)[:-1])
        with VerticalScroll(id="vertical"):
            yield Label(("SPAM\n" * 78)[:-1])
            with HorizontalScroll():
                yield Label(("v\n" * 17)[:-1])
                yield Label("@" * 302)
                yield Label("[red]>>bullseye<<[/red]", id="bullseye")
                yield Label("@" * 99)
            yield Label(("SPAM\n" * 49)[:-1])
        yield Label(("SPAM\n" * 51)[:-1])

    def key_s(self) -> None:
        self.screen.scroll_to_center(self.query_one("#bullseye"), origin_visible=False)


if __name__ == "__main__":
    MyApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Digits


class DigitApp(App):
    CSS = """
    .left {
        text-align: left;
    }
    .center {
        text-align:center;
    }
    .right {
        text-align:right;
    }
    .bold {
        text-style: bold;
    }
    """

    def compose(self) -> ComposeResult:
        yield Digits("3.14159265359", classes="left")
        yield Digits(" 0123456789+-.,ABCDEF", classes="center")
        yield Digits(" 0123456789+-.,ABCDEF", classes="center bold")
        yield Digits("3x10^4", classes="right")
        yield Digits("3x10^4", classes="right")
        yield Digits("($123.45)")
        yield Digits("£123.45")
        yield Digits("€123.45")


if __name__ == "__main__":
    app = DigitApp()
    app.run()



import pytest

from textual.app import App, ComposeResult
from textual.widgets import Label, ListItem, ListView


@pytest.mark.parametrize(
    "initial_index,expected_index",
    [
        (0, 1),
        (1, 1),
        (2, 4),
        (3, 4),
        (4, 4),
        (5, 5),
        (6, 7),
        (7, 7),
        (8, 1),
    ],
)
async def test_listview_initial_index(initial_index, expected_index) -> None:
    """Regression test for https://github.com/Textualize/textual/issues/4449"""

    class ListViewDisabledItemsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield ListView(
                ListItem(Label("0"), disabled=True),
                ListItem(Label("1")),
                ListItem(Label("2"), disabled=True),
                ListItem(Label("3"), disabled=True),
                ListItem(Label("4")),
                ListItem(Label("5")),
                ListItem(Label("6"), disabled=True),
                ListItem(Label("7")),
                ListItem(Label("8"), disabled=True),
                initial_index=initial_index,
            )

    app = ListViewDisabledItemsApp()
    async with app.run_test() as pilot:
        list_view = pilot.app.query_one(ListView)
        assert list_view.index == expected_index



import pytest
from rich.text import Text

from tests.utilities.render import render
from textual.renderables.text_opacity import TextOpacity

STOP = "\x1b[0m"


@pytest.fixture
def text():
    return Text("Hello, world!", style="#ff0000 on #00ff00", end="")


def test_simple_text_opacity(text):
    blended_red_on_green = "\x1b[38;2;127;127;0;48;2;0;255;0m"
    assert render(TextOpacity(text, opacity=0.5)) == (
        f"{blended_red_on_green}Hello, world!{STOP}"
    )


def test_value_zero_doesnt_render_the_text(text):
    assert render(TextOpacity(text, opacity=0)) == (
        f"\x1b[48;2;0;255;0m             {STOP}"
    )


def test_text_opacity_value_of_one_noop(text):
    assert render(TextOpacity(text, opacity=1)) == render(text)


def test_text_opacity_no_style_noop():
    text_no_style = Text("Hello, world!", end="")
    assert render(TextOpacity(text_no_style, opacity=0.2)) == render(text_no_style)


def test_text_opacity_only_fg_noop():
    text_only_fg = Text("Hello, world!", style="#ff0000", end="")
    assert render(TextOpacity(text_only_fg, opacity=0.5)) == render(text_only_fg)


def test_text_opacity_only_bg_noop():
    text_only_bg = Text("Hello, world!", style="on #ff0000", end="")
    assert render(TextOpacity(text_only_bg, opacity=0.5)) == render(text_only_bg)



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Static, Header, Footer


class ScreenA(Screen):
    BINDINGS = [("b", "switch_to_b", "Switch to screen B")]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static("A")
        yield Footer()

    def action_switch_to_b(self):
        self.app.switch_screen(ScreenB())


class ScreenB(Screen):
    def compose(self) -> ComposeResult:
        yield Header()
        yield Static("B")
        yield Footer()


class ModalApp(App):
    BINDINGS = [("a", "push_a", "Push screen A")]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()

    def action_push_a(self) -> None:
        self.push_screen(ScreenA())


if __name__ == "__main__":
    app = ModalApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import DataTable, Static


class HappyDataTableFunApp(App[None]):
    """The DataTable should expand to full the screen and show a horizontal scrollbar."""

    CSS = """
    #s {
        max-height: 100%;
    }
    DataTable {
        max-height: 100%;        
    }
    """

    def populate(self, table: DataTable) -> DataTable:
        for n in range(20):
            table.add_column(f"Column {n}")
        for row in range(100):
            table.add_row(*[str(row * n) for n in range(20)])
        return table

    def compose(self) -> ComposeResult:
        with Static(id="s"):
            yield self.populate(DataTable())


if __name__ == "__main__":
    HappyDataTableFunApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Placeholder


class SplitApp(App):
    CSS = """

    #split1 {
        split: right;
        width: 20;
    }

    #split2 {
        split: bottom;
        height: 10;
    }

    #split3 {
        split: top;
        height: 6;
    }

    #split4 {
        split: left;
        width: 30;
    }

    .scrollable {
        height: 5;
    }

    """

    def compose(self) -> ComposeResult:
        yield Placeholder(id="split1")
        yield Placeholder(id="split2")
        yield Placeholder(id="split3")
        yield Placeholder(id="split4")
        yield Placeholder("1", classes="scrollable")
        yield Placeholder("2", classes="scrollable")
        yield Placeholder("3", classes="scrollable")
        yield Placeholder("4", classes="scrollable")
        yield Placeholder("5", classes="scrollable")


if __name__ == "__main__":
    app = SplitApp()
    app.run()



from textual.app import App
from textual.widgets import Static


class GridApp(App):
    CSS_PATH = "grid.tcss"

    def compose(self):
        yield Static("Grid cell 1\n\nrow-span: 3;\ncolumn-span: 2;", id="static1")
        yield Static("Grid cell 2", id="static2")
        yield Static("Grid cell 3", id="static3")
        yield Static("Grid cell 4", id="static4")
        yield Static("Grid cell 5", id="static5")
        yield Static("Grid cell 6", id="static6")
        yield Static("Grid cell 7", id="static7")


if __name__ == "__main__":
    app = GridApp()
    app.run()



from pathlib import Path

from textual.app import App, ComposeResult
from textual.widgets import DirectoryTree


class DirectoryTreeApp(App[None]):

    def compose(self) -> ComposeResult:
        yield DirectoryTree(".")


async def test_change_directory_tree_path(tmpdir: Path) -> None:
    """The DirectoryTree should react to the path changing."""

    async with DirectoryTreeApp().run_test() as pilot:
        assert pilot.app.query_one(DirectoryTree).root.data.path == Path(".")
        pilot.app.query_one(DirectoryTree).path = tmpdir
        await pilot.pause()
        assert pilot.app.query_one(DirectoryTree).root.data.path == tmpdir



from __future__ import annotations

from typing import Any

from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Tree
from textual.widgets.tree import NodeID, TreeNode


class TreeApp(App[None]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.messages: list[tuple[str, NodeID]] = []

    def compose(self) -> ComposeResult:
        tree = Tree[str](label="tree")
        self._node = tree.root.add_leaf("leaf")
        tree.root.expand()
        yield tree

    @property
    def node(self) -> TreeNode[str]:
        return self._node

    @on(Tree.NodeHighlighted)
    @on(Tree.NodeSelected)
    @on(Tree.NodeCollapsed)
    @on(Tree.NodeExpanded)
    def record_event(
        self,
        event: (
            Tree.NodeHighlighted[str]
            | Tree.NodeSelected[str]
            | Tree.NodeCollapsed[str]
            | Tree.NodeExpanded[str]
        ),
    ) -> None:
        self.messages.append((event.__class__.__name__, event.node.id))


async def test_move_cursor() -> None:
    """Test moving the cursor to a node (updating the highlighted node)."""
    async with TreeApp().run_test() as pilot:
        app = pilot.app
        tree: Tree[str] = app.query_one(Tree)
        node_to_move_to = app.node
        tree.move_cursor(node_to_move_to)
        await pilot.pause()

        # Note there are no Selected messages. We only move the cursor.
        assert app.messages == [
            ("NodeExpanded", 0),  # From the call to `tree.root.expand()` in compose
            ("NodeHighlighted", 0),  # From the initial highlight of the root node
            ("NodeHighlighted", 1),  # From the call to `tree.move_cursor`
        ]


async def test_move_cursor_reset() -> None:
    async with TreeApp().run_test() as pilot:
        app = pilot.app
        tree: Tree[str] = app.query_one(Tree)
        tree.move_cursor(app.node)
        tree.move_cursor(None)
        await pilot.pause()
        assert app.messages == [
            ("NodeExpanded", 0),  # From the call to `tree.root.expand()` in compose
            ("NodeHighlighted", 0),  # From the initial highlight of the root node
            ("NodeHighlighted", 1),  # From the 1st call to `tree.move_cursor`
            ("NodeHighlighted", 0),  # From the call to `tree.move_cursor(None)`
        ]


async def test_select_node() -> None:
    async with TreeApp().run_test() as pilot:
        app = pilot.app
        tree: Tree[str] = app.query_one(Tree)
        tree.select_node(app.node)
        await pilot.pause()
        assert app.messages == [
            ("NodeExpanded", 0),  # From the call to `tree.root.expand()` in compose
            ("NodeHighlighted", 0),  # From the initial highlight of the root node
            ("NodeHighlighted", 1),  # From the `tree.select_node` call
            ("NodeSelected", 1),  # From the call to `tree.select_node`
        ]


async def test_select_node_reset() -> None:
    async with TreeApp().run_test() as pilot:
        app = pilot.app
        tree: Tree[str] = app.query_one(Tree)
        tree.move_cursor(app.node)
        tree.select_node(None)
        await pilot.pause()

        # Notice no Selected messages.
        assert app.messages == [
            ("NodeExpanded", 0),  # From the call to `tree.root.expand()` in compose
            ("NodeHighlighted", 0),  # From the initial highlight of the root node
            ("NodeHighlighted", 1),  # From the `tree.move_cursor` call
            ("NodeHighlighted", 0),  # From the call to `tree.select_node(None)`
        ]



from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widgets import Static


class HeightApp(App[None]):

    CSS = """
    Horizontal {
        border: solid red;
        height: auto;
    }

    Static {
        border: solid green;
        width: auto;
    }

    #fill_parent {
        height: 100%;
    }

    #static {
        height: 16;
    }
    """

    def compose(self) -> ComposeResult:
        yield Horizontal(
            Static("As tall as container", id="fill_parent"),
            Static("This has default\nheight\nbut a\nfew lines"),
            Static("I have a static height", id="static"),
        )


if __name__ == "__main__":
    HeightApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Checkbox, Footer


class ScrollOffByOne(App):
    """Scroll to item 50."""

    AUTO_FOCUS = None

    def compose(self) -> ComposeResult:
        for number in range(1, 100):
            yield Checkbox(str(number), id=f"number-{number}")
        yield Footer()

    def on_ready(self) -> None:
        self.query_one("#number-50").scroll_visible()


app = ScrollOffByOne()
if __name__ == "__main__":
    app.run()



from textual.app import App
from textual.command import CommandPalette, Hit, Hits, Provider
from textual.screen import Screen
from textual.system_commands import SystemCommandsProvider


async def test_sources_with_no_known_screen() -> None:
    """A command palette with no known screen should have an empty source set."""
    assert CommandPalette()._provider_classes == set()


class ExampleCommandSource(Provider):
    async def search(self, _: str) -> Hits:
        def goes_nowhere_does_nothing() -> None:
            pass

        yield Hit(1, "Hit", goes_nowhere_does_nothing, "Hit")


class AppWithActiveCommandPalette(App[None]):
    def on_mount(self) -> None:
        self.action_command_palette()


class AppWithNoSources(AppWithActiveCommandPalette):
    pass


async def test_no_app_command_sources() -> None:
    """An app with no sources declared should work fine."""
    async with AppWithNoSources().run_test() as pilot:
        assert isinstance(pilot.app.screen, CommandPalette)
        assert pilot.app.screen._provider_classes == {SystemCommandsProvider}


class AppWithSources(AppWithActiveCommandPalette):
    COMMANDS = {ExampleCommandSource}


async def test_app_command_sources() -> None:
    """Command sources declared on an app should be in the command palette."""
    async with AppWithSources().run_test() as pilot:
        assert isinstance(pilot.app.screen, CommandPalette)
        assert pilot.app.screen._provider_classes == {ExampleCommandSource}


class AppWithInitialScreen(App[None]):
    def __init__(self, screen: Screen) -> None:
        super().__init__()
        self._test_screen = screen

    def on_mount(self) -> None:
        self.push_screen(self._test_screen)


class ScreenWithNoSources(Screen[None]):
    def on_mount(self) -> None:
        self.app.action_command_palette()


async def test_no_screen_command_sources() -> None:
    """An app with a screen with no sources declared should work fine."""
    async with AppWithInitialScreen(ScreenWithNoSources()).run_test() as pilot:
        assert isinstance(pilot.app.screen, CommandPalette)
        assert pilot.app.screen._provider_classes == {SystemCommandsProvider}


class ScreenWithSources(ScreenWithNoSources):
    COMMANDS = {ExampleCommandSource}


async def test_screen_command_sources() -> None:
    """Command sources declared on a screen should be in the command palette."""
    async with AppWithInitialScreen(ScreenWithSources()).run_test() as pilot:
        assert isinstance(pilot.app.screen, CommandPalette)
        assert pilot.app.screen._provider_classes == {
            SystemCommandsProvider,
            ExampleCommandSource,
        }


class AnotherCommandSource(ExampleCommandSource):
    pass


class CombinedSourceApp(App[None]):
    COMMANDS = {AnotherCommandSource}

    def on_mount(self) -> None:
        self.push_screen(ScreenWithSources())


async def test_app_and_screen_command_sources_combine() -> None:
    """If an app and the screen have command sources they should combine."""
    async with CombinedSourceApp().run_test() as pilot:
        assert isinstance(pilot.app.screen, CommandPalette)
        assert (
            pilot.app.screen._provider_classes
            == CombinedSourceApp.COMMANDS | ScreenWithSources.COMMANDS
        )



import pytest

from textual.widgets import Rule
from textual.widgets.rule import InvalidLineStyle, InvalidRuleOrientation


def test_invalid_rule_orientation():
    with pytest.raises(InvalidRuleOrientation):
        Rule(orientation="invalid orientation!")


def test_invalid_rule_line_style():
    with pytest.raises(InvalidLineStyle):
        Rule(line_style="invalid line style!")


def test_invalid_reactive_rule_orientation_change():
    rule = Rule()
    with pytest.raises(InvalidRuleOrientation):
        rule.orientation = "invalid orientation!"


def test_invalid_reactive_rule_line_style_change():
    rule = Rule()
    with pytest.raises(InvalidLineStyle):
        rule.line_style = "invalid line style!"



from __future__ import annotations

import pytest

from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Tab, Tabs
from textual.widgets._tabs import Underline


async def test_tab_label():
    """It should be possible to access a tab's label."""
    assert Tab("Pilot").label_text == "Pilot"


async def test_tab_relabel():
    """It should be possible to relabel a tab."""
    tab = Tab("Pilot")
    assert tab.label_text == "Pilot"
    tab.label = "Aeryn"
    assert tab.label_text == "Aeryn"


async def test_compose_empty_tabs():
    """It should be possible to create an empty Tabs."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs()

    async with TabsApp().run_test() as pilot:
        assert pilot.app.query_one(Tabs).tab_count == 0
        assert pilot.app.query_one(Tabs).active_tab is None


async def test_compose_tabs_from_strings():
    """It should be possible to create a Tabs from some strings."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("John", "Aeryn", "Moya", "Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"


async def test_compose_tabs_from_tabs():
    """It should be possible to create a Tabs from some Tabs."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs(
                Tab("John"),
                Tab("Aeryn"),
                Tab("Moya"),
                Tab("Pilot"),
            )

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"


async def test_add_tabs_later():
    """It should be possible to add tabs later on in the app's cycle."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs()

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 0
        assert tabs.active_tab is None
        await tabs.add_tab("John")
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        await tabs.add_tab("Aeryn")
        assert tabs.tab_count == 2
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"


async def test_add_tab_before():
    """It should be possible to add a tab before another tab."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        await tabs.add_tab("John", before="tab-1")
        assert tabs.tab_count == 2
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        await tabs.add_tab("John", before=tabs.active_tab)
        assert tabs.tab_count == 3
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"


async def test_add_tab_before_badly():
    """Test exceptions from badly adding a tab before another."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        with pytest.raises(Tabs.TabError):
            tabs.add_tab("John", before="this-is-not-a-tab")
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        with pytest.raises(Tabs.TabError):
            tabs.add_tab("John", before=Tab("I just made this up"))
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"


async def test_add_tab_after():
    """It should be possible to add a tab after another tab."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        await tabs.add_tab("John", after="tab-1")
        assert tabs.tab_count == 2
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        await tabs.add_tab("John", after=tabs.active_tab)
        assert tabs.tab_count == 3
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"


async def test_add_tab_after_badly():
    """Test exceptions from badly adding a tab after another."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        with pytest.raises(Tabs.TabError):
            tabs.add_tab("John", after="this-is-not-a-tab")
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        with pytest.raises(Tabs.TabError):
            tabs.add_tab("John", after=Tab("I just made this up"))
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"


async def test_add_tab_before_and_after():
    """Attempting to add a tab before and after another is an error."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == "tab-1"
        with pytest.raises(Tabs.TabError):
            tabs.add_tab("John", before="tab-1", after="tab-1")


async def test_remove_tabs():
    """It should be possible to remove tabs."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("John", "Aeryn", "Moya", "Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"

        await tabs.remove_tab("tab-1")
        assert tabs.tab_count == 3
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-2"

        await tabs.remove_tab(tabs.query_one("#tab-2", Tab))
        assert tabs.tab_count == 2
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-3"

        await tabs.remove_tab("tab-does-not-exist")
        assert tabs.tab_count == 2
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-3"

        await tabs.remove_tab(None)
        assert tabs.tab_count == 2
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-3"

        await tabs.remove_tab("tab-3")
        await tabs.remove_tab("tab-4")
        assert tabs.tab_count == 0
        assert tabs.active_tab is None


async def test_remove_tabs_reversed():
    """It should be possible to remove tabs."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("John", "Aeryn", "Moya", "Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"

        await tabs.remove_tab("tab-4")
        assert tabs.tab_count == 3
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"

        await tabs.remove_tab("tab-3")
        assert tabs.tab_count == 2
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"

        await tabs.remove_tab("tab-2")
        assert tabs.tab_count == 1
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"

        await tabs.remove_tab("tab-1")
        assert tabs.tab_count == 0
        assert tabs.active_tab is None


async def test_clear_tabs():
    """It should be possible to clear all tabs."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("John", "Aeryn", "Moya", "Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        await tabs.clear()
        assert tabs.tab_count == 0
        assert tabs.active_tab is None


async def test_change_active_from_code():
    """It should be possible to change the active tab from code.."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("John", "Aeryn", "Moya", "Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == tabs.active_tab.id

        tabs.active = "tab-2"
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-2"
        assert tabs.active == tabs.active_tab.id

        tabs.active = ""
        assert tabs.active_tab is None


async def test_navigate_tabs_with_keyboard():
    """It should be possible to navigate tabs with the keyboard."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("John", "Aeryn", "Moya", "Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == tabs.active_tab.id

        await pilot.press("right")
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-2"
        assert tabs.active == tabs.active_tab.id

        await pilot.press("left")
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == tabs.active_tab.id

        await pilot.press(*(["left"] * tabs.tab_count))
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"
        assert tabs.active == tabs.active_tab.id


async def test_navigate_empty_tabs_with_keyboard():
    """It should be possible to navigate an empty tabs with the keyboard."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs()

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 0
        assert tabs.active_tab is None
        assert tabs.active == ""

        await pilot.press("right")
        assert tabs.active_tab is None
        assert tabs.active == ""

        await pilot.press("left")
        assert tabs.active_tab is None
        assert tabs.active == ""


async def test_navigate_tabs_with_mouse():
    """It should be possible to navigate tabs with the mouse."""

    class TabsApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Tabs("John", "Aeryn", "Moya", "Pilot")

    async with TabsApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        assert tabs.tab_count == 4
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"

        await pilot.click("#tab-2")
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-2"

        await pilot.click("Underline", offset=(2, 0))
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "tab-1"


class TabsMessageCatchApp(App[None]):
    def __init__(self) -> None:
        super().__init__()
        self.intended_handlers: list[str] = []

    def compose(self) -> ComposeResult:
        yield Tabs("John", "Aeryn", "Moya", "Pilot")

    @on(Tabs.Cleared)
    @on(Tabs.TabActivated)
    @on(Underline.Clicked)
    @on(Tab.Clicked)
    def log_message(
        self, event: Tabs.Cleared | Tabs.TabActivated | Underline.Clicked | Tab.Clicked
    ) -> None:
        self.intended_handlers.append(event.handler_name)

    @on(Tabs.TabActivated)
    @on(Tabs.Cleared)
    def check_control(self, event: Tabs.TabActivated) -> None:
        assert event.control is event.tabs


async def test_startup_messages():
    """On startup there should be a tab activated message."""
    async with TabsMessageCatchApp().run_test() as pilot:
        assert pilot.app.intended_handlers == ["on_tabs_tab_activated"]


async def test_change_tab_with_code_messages():
    """Changing tab in code should result in an activated tab message."""
    async with TabsMessageCatchApp().run_test() as pilot:
        pilot.app.query_one(Tabs).active = "tab-2"
        await pilot.pause()
        assert pilot.app.intended_handlers == [
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
        ]


async def test_remove_tabs_messages():
    """Removing tabs should result in various messages."""
    async with TabsMessageCatchApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        for n in range(4):
            await tabs.remove_tab(f"tab-{n+1}")

        await pilot.pause()
        assert pilot.app.intended_handlers == [
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
            "on_tabs_cleared",
        ]


async def test_reverse_remove_tabs_messages():
    """Removing tabs should result in various messages."""
    async with TabsMessageCatchApp().run_test() as pilot:
        tabs = pilot.app.query_one(Tabs)
        for n in reversed(range(4)):
            await tabs.remove_tab(f"tab-{n+1}")

        await pilot.pause()
        assert pilot.app.intended_handlers == [
            "on_tabs_tab_activated",
            "on_tabs_cleared",
        ]


async def test_keyboard_navigation_messages():
    """Keyboard navigation should result in the expected messages."""
    async with TabsMessageCatchApp().run_test() as pilot:
        await pilot.press("right")
        await pilot.pause()
        await pilot.press("left")
        await pilot.pause()
        assert pilot.app.intended_handlers == [
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
        ]


async def test_mouse_navigation_messages():
    """Mouse navigation should result in the expected messages."""
    async with TabsMessageCatchApp().run_test() as pilot:
        await pilot.click("#tab-2")
        await pilot.pause()
        await pilot.click("Underline", offset=(2, 0))
        await pilot.pause()
        assert pilot.app.intended_handlers == [
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
            "on_tabs_tab_activated",
        ]


async def test_disabled_tab_is_not_activated_by_clicking_underline():
    """Regression test for https://github.com/Textualize/textual/issues/4701"""

    class DisabledTabApp(App):
        def compose(self) -> ComposeResult:
            yield Tabs(
                Tab("Enabled", id="enabled"),
                Tab("Disabled", id="disabled", disabled=True),
            )

    app = DisabledTabApp()
    async with app.run_test() as pilot:
        # Click the underline beneath the disabled tab
        await pilot.click(Tabs, offset=(14, 2))
        tabs = pilot.app.query_one(Tabs)
        assert tabs.active_tab is not None
        assert tabs.active_tab.id == "enabled"



from __future__ import annotations

from fractions import Fraction
from typing import NamedTuple

from textual.geometry import Spacing


class BoxModel(NamedTuple):
    """The result of `get_box_model`."""

    # Content + padding + border
    width: Fraction
    height: Fraction
    margin: Spacing  # Additional margin



"""
Helper script to help document all widgets.

This goes through the widgets listed in textual.widgets and prints the scaffolding
for the tables that are used to document the classvars BINDINGS and COMPONENT_CLASSES.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import textual.widgets

if TYPE_CHECKING:
    from textual.binding import Binding


def print_bindings(widget: str, bindings: list[Binding]) -> None:
    """Print a table summarising the bindings.

    The table contains columns for the key(s) that trigger the binding,
    the method that it calls (and tries to link it to the widget itself),
    and the description of the binding.
    """
    if bindings:
        print("BINDINGS")
        print('"""')
        print("| Key(s) | Description |")
        print("| :- | :- |")

    for binding in bindings:
        print(f"| {binding.key} | {binding.description} |")

    if bindings:
        print('"""')


def print_component_classes(classes: set[str]) -> None:
    """Print a table to document these component classes.

    The table contains two columns, one with the component class name and another
    for the description of what the component class is for.
    The second column is always empty.
    """
    if classes:
        print("COMPONENT_CLASSES")
        print('"""')
        print("| Class | Description |")
        print("| :- | :- |")

    for cls in sorted(classes):
        print(f"| `{cls}` | XXX |")

    if classes:
        print('"""')


def main() -> None:
    """Main entrypoint.

    Iterates over all widgets and prints docs tables.
    """

    widgets: list[str] = textual.widgets.__all__

    for widget in widgets:
        w = getattr(textual.widgets, widget)
        bindings: list[Binding] = w.__dict__.get("BINDINGS", [])
        component_classes: set[str] = getattr(w, "COMPONENT_CLASSES", set())

        if bindings or component_classes:
            print(widget)
            print()
            print_bindings(widget, bindings)
            print_component_classes(component_classes)
            print()


if __name__ == "__main__":
    main()



import pytest

from textual._node_list import NodeList
from textual.widget import Widget


def test_empty_list():
    """Does an empty node list report as being empty?"""
    assert len(NodeList()) == 0


def test_add_one():
    """Does adding a node to the node list report as having one item?"""
    nodes = NodeList()
    nodes._append(Widget())
    assert len(nodes) == 1


def test_length_hint():
    """Check length hint dunder method."""
    nodes = NodeList()
    assert nodes.__length_hint__() == 0
    nodes._append(Widget())
    nodes._append(Widget())
    nodes._append(Widget())
    assert nodes.__length_hint__() == 3


def test_repeat_add_one():
    """Does adding the same item to the node list ignore the additional adds?"""
    nodes = NodeList()
    widget = Widget()
    for _ in range(1000):
        nodes._append(widget)
    assert len(nodes) == 1


def test_insert():
    nodes = NodeList()
    widget1 = Widget()
    widget2 = Widget()
    widget3 = Widget()
    nodes._append(widget1)
    nodes._append(widget3)
    nodes._insert(1, widget2)
    assert list(nodes) == [widget1, widget2, widget3]


def test_truthy():
    """Does a node list act as a truthy object?"""
    nodes = NodeList()
    assert not bool(nodes)
    nodes._append(Widget())
    assert bool(nodes)


def test_contains():
    """Can we check if a widget is (not) within the list?"""
    widget = Widget()
    nodes = NodeList()
    assert widget not in nodes
    nodes._append(widget)
    assert widget in nodes
    assert Widget() not in nodes


def test_index():
    """Can we get the index of a widget in the list?"""
    widget = Widget()
    nodes = NodeList()
    with pytest.raises(ValueError):
        _ = nodes.index(widget)
    nodes._append(widget)
    assert nodes.index(widget) == 0


def test_remove():
    """Can we remove a widget we've added?"""
    widget = Widget()
    nodes = NodeList()
    nodes._append(widget)
    assert widget in nodes
    nodes._remove(widget)
    assert widget not in nodes


def test_clear():
    """Can we clear the list?"""
    nodes = NodeList()
    assert len(nodes) == 0
    widgets = [Widget() for _ in range(1000)]
    for widget in widgets:
        nodes._append(widget)
    assert len(nodes) == 1000
    for widget in widgets:
        assert widget in nodes
    nodes._clear()
    assert len(nodes) == 0
    for widget in widgets:
        assert widget not in nodes


def test_listy():
    nodes = NodeList()
    widget1 = Widget()
    widget2 = Widget()
    nodes._append(widget1)
    nodes._append(widget2)
    assert list(nodes) == [widget1, widget2]
    assert list(reversed(nodes)) == [widget2, widget1]
    assert nodes[0] == widget1
    assert nodes[1] == widget2
    assert nodes[0:2] == [widget1, widget2]



from rich import inspect
from rich.console import Console

c = Console(record=True, width=110)

f = open("foo.txt", "w")

inspect(f, console=c)
c.save_svg("inspect1.svg")


inspect(f, console=c, methods=True)
c.save_svg("inspect2.svg")


inspect(f, console=c, methods=True, help=True)
c.save_svg("inspect3.svg")



from textual.app import App


class ToastApp(App[None]):
    def on_mount(self) -> None:
        # Show an information notification.
        self.notify("It's an older code, sir, but it checks out.")

        # Show a warning. Note that Textual's notification system allows
        # for the use of Rich console markup.
        self.notify(
            "Now witness the firepower of this fully "
            "[b]ARMED[/b] and [i][b]OPERATIONAL[/b][/i] battle station!",
            title="Possible trap detected",
            severity="warning",
        )

        # Show an error. Set a longer timeout so it's noticed.
        self.notify("It's a trap!", severity="error", timeout=10)

        # Show an information notification, but without any sort of title.
        self.notify("It's against my programming to impersonate a deity.", title="")


if __name__ == "__main__":
    ToastApp().run()



from textual.app import App, ComposeResult
from textual.widgets import MarkdownViewer

EXAMPLE_MARKDOWN = """\
# Markdown Viewer

This is an example of Textual's `MarkdownViewer` widget.


## Features

Markdown syntax and extensions are supported.

- Typography *emphasis*, **strong**, `inline code` etc.
- Headers
- Lists (bullet and ordered)
- Syntax highlighted code blocks
- Tables!

## Tables

Tables are displayed in a DataTable widget.

| Name            | Type   | Default | Description                        |
| --------------- | ------ | ------- | ---------------------------------- |
| `show_header`   | `bool` | `True`  | Show the table header              |
| `fixed_rows`    | `int`  | `0`     | Number of fixed rows               |
| `fixed_columns` | `int`  | `0`     | Number of fixed columns            |
| `zebra_stripes` | `bool` | `False` | Display alternating colors on rows |
| `header_height` | `int`  | `1`     | Height of header row               |
| `show_cursor`   | `bool` | `True`  | Show a cell cursor                 |


## Code Blocks

Code blocks are syntax highlighted, with guidelines.

```python
class ListViewExample(App):
    def compose(self) -> ComposeResult:
        yield ListView(
            ListItem(Label("One")),
            ListItem(Label("Two")),
            ListItem(Label("Three")),
        )
        yield Footer()
```
"""


class MarkdownExampleApp(App):
    def compose(self) -> ComposeResult:
        yield MarkdownViewer(EXAMPLE_MARKDOWN, show_table_of_contents=True)


if __name__ == "__main__":
    app = MarkdownExampleApp()
    app.run()



import pytest
from rich.segment import Segment
from rich.style import Style

from textual._segment_tools import NoCellPositionForIndex
from textual.color import Color
from textual.filter import Monochrome
from textual.strip import Strip


def test_cell_length() -> None:
    strip = Strip([Segment("foo"), Segment("💩"), Segment("bar")])
    assert strip._cell_length is None
    assert strip.cell_length == 8
    assert strip._cell_length == 8


def test_repr() -> None:
    strip = Strip([Segment("foo")])
    assert repr(strip) == "Strip([Segment('foo')], 3)"


def test_join() -> None:
    strip1 = Strip([Segment("foo")])
    strip2 = Strip([Segment("bar")])
    strip = Strip.join([strip1, strip2])
    assert len(strip) == 2
    assert strip.cell_length == 6
    assert list(strip) == [Segment("foo"), Segment("bar")]


def test_bool() -> None:
    assert not Strip([])
    assert Strip([Segment("foo")])


def test_iter() -> None:
    assert list(Strip([])) == []
    assert list(Strip([Segment("foo")])) == [Segment("foo")]
    assert list(Strip([Segment("foo"), Segment("bar")])) == [
        Segment("foo"),
        Segment("bar"),
    ]


def test_len():
    assert len(Strip([])) == 0
    assert len(Strip([Segment("foo")])) == 1
    assert len(Strip([Segment("foo"), Segment("bar")])) == 2


def test_reversed():
    assert list(reversed(Strip([]))) == []
    assert list(reversed(Strip([Segment("foo")]))) == [Segment("foo")]
    assert list(reversed(Strip([Segment("foo"), Segment("bar")]))) == [
        Segment("bar"),
        Segment("foo"),
    ]


def test_eq():
    assert Strip([]) == Strip([])
    assert Strip([Segment("foo")]) == Strip([Segment("foo")])
    assert Strip([Segment("foo")]) != Strip([Segment("bar")])


def test_adjust_cell_length():
    assert Strip([]).adjust_cell_length(3) == Strip([Segment("   ")])
    assert Strip([Segment("f")]).adjust_cell_length(3) == Strip(
        [Segment("f"), Segment("  ")]
    )
    assert Strip([Segment("💩")]).adjust_cell_length(3) == Strip(
        [Segment("💩"), Segment(" ")]
    )

    assert Strip([Segment("💩💩")]).adjust_cell_length(3) == Strip([Segment("💩 ")])
    assert Strip([Segment("💩💩")]).adjust_cell_length(4) == Strip([Segment("💩💩")])
    assert Strip([Segment("💩"), Segment("💩💩")]).adjust_cell_length(2) == Strip(
        [Segment("💩")]
    )
    assert Strip([Segment("💩"), Segment("💩💩")]).adjust_cell_length(4) == Strip(
        [Segment("💩"), Segment("💩")]
    )


def test_extend_cell_length():
    strip = Strip([Segment("foo"), Segment("bar")])
    assert strip.extend_cell_length(3).text == "foobar"
    assert strip.extend_cell_length(6).text == "foobar"
    assert strip.extend_cell_length(7).text == "foobar "
    assert strip.extend_cell_length(9).text == "foobar   "


def test_simplify():
    assert Strip([Segment("foo"), Segment("bar")]).simplify() == Strip(
        [Segment("foobar")]
    )


def test_apply_filter():
    strip = Strip([Segment("foo", Style.parse("red"))])
    expected = Strip([Segment("foo", Style.parse("#1b1b1b"))])
    assert strip.apply_filter(Monochrome(), Color(0, 0, 0)) == expected


def test_style_links():
    link_style = Style.on(click="clicked")
    strip = Strip(
        [
            Segment("foo"),
            Segment("bar", link_style),
            Segment("baz"),
        ]
    )
    hover_style = Style(underline=True)
    new_strip = strip.style_links(link_style._link_id, hover_style)
    expected = Strip(
        [
            Segment("foo"),
            Segment("bar", link_style + hover_style),
            Segment("baz"),
        ]
    )
    assert new_strip == expected


def test_crop():
    assert Strip([Segment("foo")]).crop(0, 3) == Strip([Segment("foo")])
    assert Strip([Segment("foo")]).crop(0, 2) == Strip([Segment("fo")])
    assert Strip([Segment("foo")]).crop(0, 1) == Strip([Segment("f")])

    assert Strip([Segment("foo")]).crop(1, 3) == Strip([Segment("oo")])
    assert Strip([Segment("foo")]).crop(1, 2) == Strip([Segment("o")])
    assert Strip([Segment("foo")]).crop(1, 1) == Strip([])

    assert Strip([Segment("foo💩"), Segment("b💩ar"), Segment("ba💩z")]).crop(
        1, 6
    ) == Strip([Segment("oo💩"), Segment("b")])


@pytest.mark.parametrize(
    "text,crop,output",
    [
        ["foo", (0, 5), [Segment("foo")]],
        ["foo", (2, 5), [Segment("o")]],
        ["foo", (3, 5), []],
        ["foo", (4, 6), []],
    ],
)
def test_crop_out_of_bounds(text, crop, output):
    assert Strip([Segment(text)]).crop(*crop) == Strip(output)


def test_divide():
    assert Strip([Segment("foo")]).divide([1, 2]) == [
        Strip([Segment("f")]),
        Strip([Segment("o")]),
    ]


@pytest.mark.parametrize(
    "index,cell_position",
    [
        (0, 0),
        (1, 1),
        (2, 2),
        (3, 3),
        (4, 4),
        (5, 6),
        (6, 8),
        (7, 10),
        (8, 11),
        (9, 12),
        (10, 13),
        (11, 14),
    ],
)
def test_index_to_cell_position(index, cell_position):
    strip = Strip([Segment("ab"), Segment("cd日本語ef"), Segment("gh")])
    assert cell_position == strip.index_to_cell_position(index)


def test_index_cell_position_no_segments():
    strip = Strip([])
    with pytest.raises(NoCellPositionForIndex):
        strip.index_to_cell_position(2)


def test_index_cell_position_index_too_large():
    strip = Strip([Segment("abcdef"), Segment("ghi")])
    with pytest.raises(NoCellPositionForIndex):
        strip.index_to_cell_position(100)


def test_text():
    assert Strip([]).text == ""
    assert Strip([Segment("foo")]).text == "foo"
    assert Strip([Segment("foo"), Segment("bar")]).text == "foobar"



from __future__ import annotations

import random
from datetime import timedelta
from time import monotonic

from textual import events
from textual.app import App, ComposeResult
from textual.containers import Grid
from textual.reactive import var
from textual.renderables.gradient import LinearGradient
from textual.widget import Widget
from textual.widgets import Digits, Label, Switch

# A nice rainbow of colors.
COLORS = [
    "#881177",
    "#aa3355",
    "#cc6666",
    "#ee9944",
    "#eedd00",
    "#99dd55",
    "#44dd88",
    "#22ccbb",
    "#00bbcc",
    "#0099cc",
    "#3366bb",
    "#663399",
]


# Maps a switch number on to other switch numbers, which should be toggled.
TOGGLES: dict[int, tuple[int, ...]] = {
    1: (2, 4, 5),
    2: (1, 3),
    3: (2, 5, 6),
    4: (1, 7),
    5: (2, 4, 6, 8),
    6: (3, 9),
    7: (4, 5, 8),
    8: (7, 9),
    9: (5, 6, 8),
}


class LabelSwitch(Widget):
    """Switch with a numeric label."""

    DEFAULT_CSS = """
    LabelSwitch Label {
        text-align: center;
        width: 1fr;
        text-style: bold;
    }

    LabelSwitch Label#label-5 {
        color: $text-disabled;
    }
    """

    def __init__(self, switch_no: int) -> None:
        self.switch_no = switch_no
        super().__init__()

    def compose(self) -> ComposeResult:
        """Compose the label and a switch."""
        yield Label(str(self.switch_no), id=f"label-{self.switch_no}")
        yield Switch(id=f"switch-{self.switch_no}", name=str(self.switch_no))


class Timer(Digits):
    """Displays a timer that stops when you win."""

    DEFAULT_CSS = """
    Timer {
        text-align: center;
        width: auto;
        margin: 2 8;
        color: $warning;
        &:light {
            color: $secondary;
        }
    }
    """
    start_time = var(0.0)
    running = var(True)

    def on_mount(self) -> None:
        """Start the timer on mount."""
        self.start_time = monotonic()
        self.set_interval(1, self.tick)
        self.tick()

    def tick(self) -> None:
        """Called from `set_interval` to update the clock."""
        if self.start_time == 0 or not self.running:
            return
        time_elapsed = timedelta(seconds=int(monotonic() - self.start_time))
        self.update(str(time_elapsed))


class MerlinApp(App):
    """A simple reproduction of one game on the Merlin hand held console."""

    CSS = """
    Screen {
        align: center middle;
    }

    Screen.-win {
        background: transparent;
    }

    Screen.-win Timer {
        color: $success;
    }

    Grid {
        width: auto;
        height: auto;
        border: thick $border;
        padding: 1 2;
        grid-size: 3 3;
        grid-rows: auto;
        grid-columns: auto;
        grid-gutter: 1 1;
        background: $surface;
    }
    """

    def render(self) -> LinearGradient:
        """Renders a gradient, when the background is transparent."""
        stops = [(i / (len(COLORS) - 1), c) for i, c in enumerate(COLORS)]
        return LinearGradient(30.0, stops)

    def compose(self) -> ComposeResult:
        """Compose a timer, and a grid of 9 switches."""
        yield Timer()
        with Grid():
            for switch in (7, 8, 9, 4, 5, 6, 1, 2, 3):
                yield LabelSwitch(switch)

    def on_mount(self) -> None:
        """Randomize the switches on mount."""
        for switch_no in range(1, 10):
            if random.randint(0, 1):
                self.query_one(f"#switch-{switch_no}", Switch).toggle()

    def check_win(self) -> bool:
        """Check for a win."""
        on_switches = {
            int(switch.name or "0") for switch in self.query(Switch) if switch.value
        }
        return on_switches == {1, 2, 3, 4, 6, 7, 8, 9}

    def on_switch_changed(self, event: Switch.Changed) -> None:
        """Called when a switch is toggled."""
        # The switch that was pressed
        switch_no = int(event.switch.name or "0")
        # Also toggle corresponding switches
        with self.prevent(Switch.Changed):
            for toggle_no in TOGGLES[switch_no]:
                self.query_one(f"#switch-{toggle_no}", Switch).toggle()
        # Check the win
        if self.check_win():
            self.query_one("Screen").add_class("-win")
            self.query_one(Timer).running = False
            self.notify("You win!", title="congratulations", severity="information")

    def on_key(self, event: events.Key) -> None:
        """Maps switches to keys, so we can use the keyboard as well."""
        if event.character and event.character.isdigit():
            self.query_one(f"#switch-{event.character}", Switch).toggle()


if __name__ == "__main__":
    MerlinApp().run()



"""Provides a simple Label widget."""

from __future__ import annotations

from typing import Literal

from rich.console import RenderableType

from textual.visual import SupportsVisual
from textual.widgets._static import Static

LabelVariant = Literal["success", "error", "warning", "primary", "secondary", "accent"]


class Label(Static):
    """A simple label widget for displaying text-oriented renderables."""

    DEFAULT_CSS = """
    Label {
        width: auto;
        height: auto;
        min-height: 1;

        &.success {
            color: $text-success;
            background: $success-muted;
        }
        &.error {
            color: $text-error;
            background: $error-muted;
        }
        &.warning {
            color: $text-warning;
            background: $warning-muted;
        }
        &.primary {
            color: $text-primary;
            background: $primary-muted;
        }
        &.secondary {
            color: $text-secondary;
            background: $secondary-muted;
        }
        &.accent {
            color: $text-accent;
            background: $accent-muted;
        }
    }
    """

    def __init__(
        self,
        renderable: RenderableType | SupportsVisual = "",
        *,
        variant: LabelVariant | None = None,
        expand: bool = False,
        shrink: bool = False,
        markup: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            renderable,
            expand=expand,
            shrink=shrink,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        if variant:
            self.add_class(variant)



from __future__ import annotations

from typing import NamedTuple

import pytest

from textual._layout_resolve import layout_resolve


class Edge(NamedTuple):
    size: int | None = None
    fraction: int = 1
    min_size: int = 1


def test_empty():
    assert layout_resolve(10, []) == []


def test_total_zero():
    assert layout_resolve(0, [Edge(10)]) == [10]


def test_single():
    # One edge fixed size
    assert layout_resolve(100, [Edge(10)]) == [10]
    # One edge fraction of 1
    assert layout_resolve(100, [Edge(None, 1)]) == [100]
    # One edge fraction 3
    assert layout_resolve(100, [Edge(None, 2)]) == [100]
    # One edge, fraction1, min size 20
    assert layout_resolve(100, [Edge(None, 1, 20)]) == [100]
    # One edge fraction 1, min size 120
    assert layout_resolve(100, [Edge(None, 1, 120)]) == [120]


def test_two():
    # Two edges fixed size
    assert layout_resolve(100, [Edge(10), Edge(20)]) == [10, 20]
    # Two edges, fixed size of one exceeds total
    assert layout_resolve(100, [Edge(120), Edge(None, 1)]) == [120, 1]
    # Two edges, fraction 1 each
    assert layout_resolve(100, [Edge(None, 1), Edge(None, 1)]) == [50, 50]
    # Two edges, one with fraction 2, one with fraction 1
    # Note first value is rounded down, second is rounded up
    assert layout_resolve(100, [Edge(None, 2), Edge(None, 1)]) == [66, 34]
    # Two edges, both with fraction 2
    assert layout_resolve(100, [Edge(None, 2), Edge(None, 2)]) == [50, 50]
    # Two edges, one with fraction 3, one with fraction 1
    assert layout_resolve(100, [Edge(None, 3), Edge(None, 1)]) == [75, 25]
    # Two edges, one with fraction 3, one with fraction 1, second with min size of 30
    assert layout_resolve(100, [Edge(None, 3), Edge(None, 1, 30)]) == [70, 30]
    # Two edges, one with fraction 1 and min size 30, one with fraction 3
    assert layout_resolve(100, [Edge(None, 1, 30), Edge(None, 3)]) == [30, 70]


@pytest.mark.parametrize(
    "size, edges, result",
    [
        (10, [Edge(8), Edge(None, 0, 2), Edge(4)], [8, 2, 4]),
        (10, [Edge(None, 1), Edge(None, 1), Edge(None, 1)], [3, 3, 4]),
        (10, [Edge(5), Edge(None, 1), Edge(None, 1)], [5, 2, 3]),
        (10, [Edge(None, 2), Edge(None, 1), Edge(None, 1)], [5, 2, 3]),
        (10, [Edge(None, 2), Edge(3), Edge(None, 1)], [4, 3, 3]),
        (
            10,
            [Edge(None, 2), Edge(None, 1), Edge(None, 1), Edge(None, 1)],
            [4, 2, 2, 2],
        ),
        (
            10,
            [Edge(None, 4), Edge(None, 1), Edge(None, 1), Edge(None, 1)],
            [5, 2, 1, 2],
        ),
        (2, [Edge(None, 1), Edge(None, 1), Edge(None, 1)], [1, 1, 1]),
        (
            2,
            [
                Edge(None, 1, min_size=5),
                Edge(None, 1, min_size=4),
                Edge(None, 1, min_size=3),
            ],
            [5, 4, 3],
        ),
        (
            18,
            [
                Edge(None, 1, min_size=1),
                Edge(3),
                Edge(None, 1, min_size=1),
                Edge(4),
                Edge(None, 1, min_size=1),
                Edge(5),
                Edge(None, 1, min_size=1),
            ],
            [1, 3, 2, 4, 1, 5, 2],
        ),
    ],
)
def test_multiple(size, edges, result):
    assert layout_resolve(size, edges) == result



from textual.app import App
from textual.containers import Horizontal
from textual.widgets import Placeholder


class MaxHeightApp(App):
    CSS_PATH = "max_height.tcss"

    def compose(self):
        yield Horizontal(
            Placeholder("max-height: 10w", id="p1"),
            Placeholder("max-height: 999", id="p2"),
            Placeholder("max-height: 50%", id="p3"),
            Placeholder("max-height: 10", id="p4"),
        )


if __name__ == "__main__":
    app = MaxHeightApp()
    app.run()



from textual.app import App
from textual.widgets import Label


class BorderApp(App):
    CSS_PATH = "border.tcss"

    def compose(self):
        yield Label("My border is solid red", id="label1")
        yield Label("My border is dashed green", id="label2")
        yield Label("My border is tall blue", id="label3")


if __name__ == "__main__":
    app = BorderApp()
    app.run()



from rich.table import Table

from textual.app import App, ComposeResult
from textual.widgets import Static


class FizzBuzz(Static):
    def on_mount(self) -> None:
        table = Table("Number", "Fizz?", "Buzz?")
        for n in range(1, 16):
            fizz = not n % 3
            buzz = not n % 5
            table.add_row(
                str(n),
                "fizz" if fizz else "",
                "buzz" if buzz else "",
            )
        self.update(table)


class FizzBuzzApp(App):
    CSS_PATH = "fizzbuzz01.tcss"

    def compose(self) -> ComposeResult:
        yield FizzBuzz()


if __name__ == "__main__":
    app = FizzBuzzApp()
    app.run()



from textual.app import App
from textual.widgets import Label


class BackgroundApp(App):
    CSS_PATH = "background.tcss"

    def compose(self):
        yield Label("Widget 1", id="static1")
        yield Label("Widget 2", id="static2")
        yield Label("Widget 3", id="static3")


if __name__ == "__main__":
    app = BackgroundApp()
    app.run()



from textual.app import App
from textual.widgets import Label


class LinkColorApp(App):
    CSS_PATH = "link_color.tcss"

    def compose(self):
        yield Label(
            "Visit the [link='https://textualize.io']Textualize[/link] website.",
            id="lbl1",  # (1)!
        )
        yield Label(
            "Click [@click=app.bell]here[/] for the bell sound.",
            id="lbl2",  # (2)!
        )
        yield Label(
            "You can also click [@click=app.bell]here[/] for the bell sound.",
            id="lbl3",  # (3)!
        )
        yield Label(
            "[@click=app.quit]Exit this application.[/]",
            id="lbl4",  # (4)!
        )


if __name__ == "__main__":
    app = LinkColorApp()
    app.run()



from textual.app import App
from textual.widgets import Label


class AllContentAlignApp(App):
    CSS_PATH = "content_align_all.tcss"

    def compose(self):
        yield Label("left top", id="left-top")
        yield Label("center top", id="center-top")
        yield Label("right top", id="right-top")
        yield Label("left middle", id="left-middle")
        yield Label("center middle", id="center-middle")
        yield Label("right middle", id="right-middle")
        yield Label("left bottom", id="left-bottom")
        yield Label("center bottom", id="center-bottom")
        yield Label("right bottom", id="right-bottom")


if __name__ == "__main__":
    app = AllContentAlignApp()
    app.run()



"""
Utilities to move index-based selections backward/forward.

These utilities concern themselves with selections where not all options are available,
otherwise it would be enough to increment/decrement the index and use the operator `%`
to implement wrapping.
"""

from __future__ import annotations

from itertools import count
from typing import Literal, Protocol, Sequence

from typing_extensions import TypeAlias

from textual._loop import loop_from_index


class Disableable(Protocol):
    """Non-widgets that have an enabled/disabled status."""

    disabled: bool


Direction: TypeAlias = Literal[-1, 1]
"""Valid values to determine navigation direction.

In a vertical setting, 1 points down and -1 points up.
In a horizontal setting, 1 points right and -1 points left.
"""


def get_directed_distance(
    index: int, start: int, direction: Direction, wrap_at: int
) -> int:
    """Computes the distance going from `start` to `index` in the given direction.

    Starting at `start`, this is the number of steps you need to take in the given
    `direction` to reach `index`, assuming there is wrapping at 0 and `wrap_at`.
    This is also the smallest non-negative integer solution `d` to
    `(start + d * direction) % wrap_at == index`.

    The diagram below illustrates the computation of `d1 = distance(2, 8, 1, 10)` and
    `d2 = distance(2, 8, -1, 10)`:

    ```
    start ────────────────────┐
    index ────────┐           │
    indices   0 1 2 3 4 5 6 7 8 9
    d1        2 3 4           0 1
              > > >           > > (direction == 1)
    d2            6 5 4 3 2 1 0
                  < < < < < < <   (direction == -1)
    ```

    Args:
        index: The index that we want to reach.
        start: The starting point to consider when computing the distance.
        direction: The direction in which we want to compute the distance.
        wrap_at: Controls at what point wrapping around takes place.

    Returns:
        The computed distance.
    """
    return direction * (index - start) % wrap_at


def find_first_enabled(
    candidates: Sequence[Disableable],
) -> int | None:
    """Find the first enabled candidate in a sequence of possibly-disabled objects.

    Args:
        candidates: The sequence of candidates to consider.

    Returns:
        The first enabled candidate or `None` if none were available.
    """
    return next(
        (index for index, candidate in enumerate(candidates) if not candidate.disabled),
        None,
    )


def find_last_enabled(candidates: Sequence[Disableable]) -> int | None:
    """Find the last enabled candidate in a sequence of possibly-disabled objects.

    Args:
        candidates: The sequence of candidates to consider.

    Returns:
        The last enabled candidate or `None` if none were available.
    """
    total_candidates = len(candidates)
    return next(
        (
            total_candidates - offset_from_end
            for offset_from_end, candidate in enumerate(reversed(candidates), start=1)
            if not candidate.disabled
        ),
        None,
    )


def find_next_enabled(
    candidates: Sequence[Disableable],
    anchor: int | None,
    direction: Direction,
) -> int | None:
    """Find the next enabled object if we're currently at the given anchor.

    The definition of "next" depends on the given direction and this function will wrap
    around the ends of the sequence of object candidates.

    Args:
        candidates: The sequence of object candidates to consider.
        anchor: The point of the sequence from which we'll start looking for the next
            enabled object.
        direction: The direction in which to traverse the candidates when looking for
            the next enabled candidate.

    Returns:
        The next enabled object. If none are available, return the anchor.
    """

    if anchor is None:
        if candidates:
            return (
                find_first_enabled(candidates)
                if direction == 1
                else find_last_enabled(candidates)
            )
        return None

    for index, candidate in loop_from_index(candidates, anchor, direction, wrap=True):
        if not candidate.disabled:
            return index
    return anchor


def find_next_enabled_no_wrap(
    candidates: Sequence[Disableable],
    anchor: int | None,
    direction: Direction,
    with_anchor: bool = False,
) -> int | None:
    """Find the next enabled object starting from the given anchor (without wrapping).

    The meaning of "next" and "past" depend on the direction specified.

    Args:
        candidates: The sequence of object candidates to consider.
        anchor: The point of the sequence from which we'll start looking for the next
            enabled object.
        direction: The direction in which to traverse the candidates when looking for
            the next enabled candidate.
        with_anchor: Whether to consider the anchor or not.

    Returns:
        The next enabled object. If none are available, return None.
    """

    if anchor is None:
        if candidates:
            return (
                find_first_enabled(candidates)
                if direction == 1
                else find_last_enabled(candidates)
            )
        return None

    start = anchor if with_anchor else anchor + direction
    counter = count(start, direction)
    valid_candidates = (
        candidates[start:] if direction == 1 else reversed(candidates[: start + 1])
    )

    for idx, candidate in zip(counter, valid_candidates):
        if candidate.disabled:
            continue
        return idx
    return None



from __future__ import annotations

from textual.app import App, ComposeResult
from textual.widgets import Tree


class TreeApp(App[None]):
    """Test tree app."""

    def compose(self) -> ComposeResult:
        yield Tree("Test")

    def on_mount(self) -> None:
        tree = self.query_one(Tree)
        for n in range(10):
            tree.root.add(f"Trunk {n}")
        node = tree.root.children[0]
        for n in range(10):
            node = node.add(str(n))


async def test_tree_node_expand() -> None:
    """Expanding one node should not expand all nodes."""
    async with TreeApp().run_test() as pilot:
        pilot.app.query_one(Tree).root.expand()
        assert pilot.app.query_one(Tree).root.is_expanded is True
        check_node = pilot.app.query_one(Tree).root.children[0]
        while check_node.children:
            assert any(child.is_expanded for child in check_node.children) is False
            check_node = check_node.children[0]


async def test_tree_node_expand_all() -> None:
    """Expanding all on a node should expand all child nodes too."""
    async with TreeApp().run_test() as pilot:
        pilot.app.query_one(Tree).root.expand_all()
        assert pilot.app.query_one(Tree).root.is_expanded is True
        check_node = pilot.app.query_one(Tree).root.children[0]
        while check_node.children:
            assert check_node.children[0].is_expanded is True
            assert any(child.is_expanded for child in check_node.children[1:]) is False
            check_node = check_node.children[0]


async def test_tree_node_collapse() -> None:
    """Collapsing one node should not collapse all nodes."""
    async with TreeApp().run_test() as pilot:
        pilot.app.query_one(Tree).root.expand_all()
        pilot.app.query_one(Tree).root.children[0].collapse()
        assert pilot.app.query_one(Tree).root.children[0].is_expanded is False
        check_node = pilot.app.query_one(Tree).root.children[0].children[0]
        while check_node.children:
            assert all(child.is_expanded for child in check_node.children) is True
            check_node = check_node.children[0]


async def test_tree_node_collapse_all() -> None:
    """Collapsing all on a node should collapse all child noes too."""
    async with TreeApp().run_test() as pilot:
        pilot.app.query_one(Tree).root.expand_all()
        pilot.app.query_one(Tree).root.children[0].collapse_all()
        assert pilot.app.query_one(Tree).root.children[0].is_expanded is False
        check_node = pilot.app.query_one(Tree).root.children[0].children[0]
        while check_node.children:
            assert check_node.children[0].is_expanded is False
            assert all(child.is_expanded for child in check_node.children[1:]) is True
            check_node = check_node.children[0]


async def test_tree_node_toggle() -> None:
    """Toggling one node should not toggle all nodes."""
    async with TreeApp().run_test() as pilot:
        assert pilot.app.query_one(Tree).root.is_expanded is False
        check_node = pilot.app.query_one(Tree).root.children[0]
        while check_node.children:
            assert any(child.is_expanded for child in check_node.children) is False
            check_node = check_node.children[0]
        pilot.app.query_one(Tree).root.toggle()
        assert pilot.app.query_one(Tree).root.is_expanded is True
        check_node = pilot.app.query_one(Tree).root.children[0]
        while check_node.children:
            assert any(child.is_expanded for child in check_node.children) is False
            check_node = check_node.children[0]


async def test_tree_node_toggle_all() -> None:
    """Toggling all on a node should toggle all child nodes too."""
    async with TreeApp().run_test() as pilot:
        assert pilot.app.query_one(Tree).root.is_expanded is False
        check_node = pilot.app.query_one(Tree).root.children[0]
        while check_node.children:
            assert any(child.is_expanded for child in check_node.children) is False
            check_node = check_node.children[0]
        pilot.app.query_one(Tree).root.toggle_all()
        assert pilot.app.query_one(Tree).root.is_expanded is True
        check_node = pilot.app.query_one(Tree).root.children[0]
        while check_node.children:
            assert check_node.children[0].is_expanded is True
            assert any(child.is_expanded for child in check_node.children[1:]) is False
            check_node = check_node.children[0]
        pilot.app.query_one(Tree).root.toggle_all()
        assert pilot.app.query_one(Tree).root.is_expanded is False
        check_node = pilot.app.query_one(Tree).root.children[0]
        while check_node.children:
            assert any(child.is_expanded for child in check_node.children) is False
            check_node = check_node.children[0]



"""Test replacing options prompt from an option list."""
import pytest

from textual.app import App, ComposeResult
from textual.widgets import OptionList
from textual.widgets.option_list import Option, OptionDoesNotExist


class OptionListApp(App[None]):
    """Test option list application."""

    def compose(self) -> ComposeResult:
        yield OptionList(
            Option("0", id="0"),
            Option("line1\nline2"),
        )


async def test_replace_option_prompt_with_invalid_id() -> None:
    """Attempting to replace the prompt of an option ID that doesn't exist should raise an exception."""
    async with OptionListApp().run_test() as pilot:
        with pytest.raises(OptionDoesNotExist):
            pilot.app.query_one(OptionList).replace_option_prompt("does-not-exist", "new-prompt")


async def test_replace_option_prompt_with_invalid_index() -> None:
    """Attempting to replace the prompt of an option index that doesn't exist should raise an exception."""
    async with OptionListApp().run_test() as pilot:
        with pytest.raises(OptionDoesNotExist):
            pilot.app.query_one(OptionList).replace_option_prompt_at_index(23, "new-prompt")


async def test_replace_option_prompt_with_valid_id() -> None:
    """It should be possible to replace the prompt of an option ID that does exist."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        option_list.replace_option_prompt("0", "new-prompt")
        assert option_list.get_option("0").prompt == "new-prompt"


async def test_replace_option_prompt_with_valid_index() -> None:
    """It should be possible to replace the prompt of an option index that does exist."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList).replace_option_prompt_at_index(1, "new-prompt")
        assert option_list.get_option_at_index(1).prompt == "new-prompt"


async def test_replace_single_line_option_prompt_with_multiple() -> None:
    """It should be possible to replace single line prompt with multiple lines """
    new_prompt = "new-prompt\nsecond line"
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        option_list.replace_option_prompt("0", new_prompt)
        assert option_list.get_option("0").prompt == new_prompt


async def test_replace_multiple_line_option_prompt_with_single() -> None:
    """It should be possible to replace multiple line prompt with a single line"""
    new_prompt = "new-prompt"
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        option_list.replace_option_prompt("0", new_prompt)
        assert option_list.get_option("0").prompt == new_prompt


async def test_replace_multiple_line_option_prompt_with_multiple() -> None:
    """It should be possible to replace multiple line prompt with multiple lines"""
    new_prompt = "new-prompt\nsecond line"
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        option_list.replace_option_prompt_at_index(1, new_prompt)
        assert option_list.get_option_at_index(1).prompt == new_prompt



import pytest

from textual.widgets.text_area import Document, EditResult

TEXT = """I must not fear.
Fear is the mind-killer.
I forgot the rest of the quote.
Sorry Will."""


@pytest.fixture
def document():
    document = Document(TEXT)
    return document


def test_delete_single_character(document):
    replace_result = document.replace_range((0, 0), (0, 1), "")
    assert replace_result == EditResult(end_location=(0, 0), replaced_text="I")
    assert document.lines == [
        " must not fear.",
        "Fear is the mind-killer.",
        "I forgot the rest of the quote.",
        "Sorry Will.",
    ]


def test_delete_single_newline(document):
    """Testing deleting newline from right to left"""
    replace_result = document.replace_range((1, 0), (0, 16), "")
    assert replace_result == EditResult(end_location=(0, 16), replaced_text="\n")
    assert document.lines == [
        "I must not fear.Fear is the mind-killer.",
        "I forgot the rest of the quote.",
        "Sorry Will.",
    ]


def test_delete_near_end_of_document(document):
    """Test deleting a range near the end of a document."""
    replace_result = document.replace_range((1, 0), (3, 11), "")
    assert replace_result == EditResult(
        end_location=(1, 0),
        replaced_text="Fear is the mind-killer.\n"
        "I forgot the rest of the quote.\n"
        "Sorry Will.",
    )
    assert document.lines == [
        "I must not fear.",
        "",
    ]


def test_delete_clearing_the_document(document):
    replace_result = document.replace_range((0, 0), (4, 0), "")
    assert replace_result == EditResult(
        end_location=(0, 0),
        replaced_text=TEXT,
    )
    assert document.lines == [""]


def test_delete_multiple_characters_on_one_line(document):
    replace_result = document.replace_range((0, 2), (0, 7), "")
    assert replace_result == EditResult(
        end_location=(0, 2),
        replaced_text="must ",
    )
    assert document.lines == [
        "I not fear.",
        "Fear is the mind-killer.",
        "I forgot the rest of the quote.",
        "Sorry Will.",
    ]


def test_delete_multiple_lines_partially_spanned(document):
    """Deleting a selection that partially spans the first and final lines of the selection."""
    replace_result = document.replace_range((0, 2), (2, 2), "")
    assert replace_result == EditResult(
        end_location=(0, 2),
        replaced_text="must not fear.\nFear is the mind-killer.\nI ",
    )
    assert document.lines == [
        "I forgot the rest of the quote.",
        "Sorry Will.",
    ]


def test_delete_end_of_line(document):
    """Testing deleting newline from left to right"""
    replace_result = document.replace_range((0, 16), (1, 0), "")
    assert replace_result == EditResult(
        end_location=(0, 16),
        replaced_text="\n",
    )
    assert document.lines == [
        "I must not fear.Fear is the mind-killer.",
        "I forgot the rest of the quote.",
        "Sorry Will.",
    ]


def test_delete_single_line_excluding_newline(document):
    """Delete from the start to the end of the line."""
    replace_result = document.replace_range((2, 0), (2, 31), "")
    assert replace_result == EditResult(
        end_location=(2, 0),
        replaced_text="I forgot the rest of the quote.",
    )
    assert document.lines == [
        "I must not fear.",
        "Fear is the mind-killer.",
        "",
        "Sorry Will.",
    ]


def test_delete_single_line_including_newline(document):
    """Delete from the start of a line to the start of the line below."""
    replace_result = document.replace_range((2, 0), (3, 0), "")
    assert replace_result == EditResult(
        end_location=(2, 0),
        replaced_text="I forgot the rest of the quote.\n",
    )
    assert document.lines == [
        "I must not fear.",
        "Fear is the mind-killer.",
        "Sorry Will.",
    ]


TEXT_NEWLINE_EOF = """\
I must not fear.
Fear is the mind-killer.
"""


def test_delete_end_of_file_newline():
    document = Document(TEXT_NEWLINE_EOF)
    replace_result = document.replace_range((2, 0), (1, 24), "")
    assert replace_result == EditResult(end_location=(1, 24), replaced_text="\n")
    assert document.lines == [
        "I must not fear.",
        "Fear is the mind-killer.",
    ]



import pytest

from textual.suggestions import get_suggestion, get_suggestions


@pytest.mark.parametrize(
    "word, possible_words, expected_result",
    (
        ["background", ("background",), "background"],
        ["backgroundu", ("background",), "background"],
        ["bkgrund", ("background",), "background"],
        ["llow", ("background",), None],
        ["llow", ("background", "yellow"), "yellow"],
        ["yllow", ("background", "yellow", "ellow"), "yellow"],
    ),
)
def test_get_suggestion(word, possible_words, expected_result):
    assert get_suggestion(word, possible_words) == expected_result


@pytest.mark.parametrize(
    "word, possible_words, count, expected_result",
    (
        ["background", ("background",), 1, ["background"]],
        ["backgroundu", ("background",), 1, ["background"]],
        ["bkgrund", ("background",), 1, ["background"]],
        ["llow", ("background",), 1, []],
        ["llow", ("background", "yellow"), 1, ["yellow"]],
        ["yllow", ("background", "yellow", "ellow"), 1, ["yellow"]],
        ["yllow", ("background", "yellow", "ellow"), 2, ["yellow", "ellow"]],
        ["yllow", ("background", "yellow", "red"), 2, ["yellow"]],
    ),
)
def test_get_suggestions(word, possible_words, count, expected_result):
    assert get_suggestions(word, possible_words, count) == expected_result



from __future__ import annotations

import asyncio

import pytest

from textual.app import App, ComposeResult
from textual.message import Message
from textual.message_pump import MessagePump
from textual.reactive import Reactive, TooManyComputesError, reactive, var
from textual.widget import Widget

OLD_VALUE = 5_000
NEW_VALUE = 1_000_000


async def test_watch():
    """Test that changes to a watched reactive attribute happen immediately."""

    class WatchApp(App):
        count = reactive(0, init=False)

        watcher_call_count = 0

        def watch_count(self, value: int) -> None:
            self.watcher_call_count = value

    app = WatchApp()
    async with app.run_test():
        app.count += 1
        assert app.watcher_call_count == 1
        app.count += 1
        assert app.watcher_call_count == 2
        app.count -= 1
        assert app.watcher_call_count == 1
        app.count -= 1
        assert app.watcher_call_count == 0


async def test_watch_async_init_false():
    """Ensure that async watchers are called eventually when set by user code"""

    class WatchAsyncApp(App):
        count = reactive(OLD_VALUE, init=False)
        watcher_old_value = None
        watcher_new_value = None
        watcher_called_event = asyncio.Event()

        async def watch_count(self, old_value: int, new_value: int) -> None:
            self.watcher_old_value = old_value
            self.watcher_new_value = new_value
            self.watcher_called_event.set()

    app = WatchAsyncApp()
    async with app.run_test():
        app.count = NEW_VALUE
        assert app.count == NEW_VALUE  # Value is set immediately
        try:
            await asyncio.wait_for(app.watcher_called_event.wait(), timeout=0.05)
        except TimeoutError:
            pytest.fail("Async watch method (watch_count) wasn't called within timeout")

        assert app.count == NEW_VALUE  # Sanity check
        assert app.watcher_old_value == OLD_VALUE  # old_value passed to watch method
        assert app.watcher_new_value == NEW_VALUE  # new_value passed to watch method


async def test_watch_async_init_true():
    """Ensure that when init is True in a reactive, its async watcher gets called
    by Textual eventually, even when the user does not set the value themselves."""

    class WatchAsyncApp(App):
        count = reactive(OLD_VALUE, init=True)
        watcher_called_event = asyncio.Event()
        watcher_old_value = None
        watcher_new_value = None

        async def watch_count(self, old_value: int, new_value: int) -> None:
            self.watcher_old_value = old_value
            self.watcher_new_value = new_value
            self.watcher_called_event.set()

    app = WatchAsyncApp()
    async with app.run_test():
        try:
            await asyncio.wait_for(app.watcher_called_event.wait(), timeout=0.05)
        except TimeoutError:
            pytest.fail(
                "Async watcher wasn't called within timeout when reactive init = True"
            )

    assert app.count == OLD_VALUE
    assert app.watcher_old_value == OLD_VALUE
    assert app.watcher_new_value == OLD_VALUE  # The value wasn't changed


async def test_watch_init_false_always_update_false():
    class WatcherInitFalse(App):
        count = reactive(0, init=False)
        watcher_call_count = 0

        def watch_count(self, new_value: int) -> None:
            self.watcher_call_count += 1

    app = WatcherInitFalse()
    async with app.run_test():
        app.count = 0  # Value hasn't changed, and always_update=False, so watch_count shouldn't run
        assert app.watcher_call_count == 0
        app.count = 0
        assert app.watcher_call_count == 0
        app.count = 1
        assert app.watcher_call_count == 1


async def test_watch_init_true():
    class WatcherInitTrue(App):
        count = var(OLD_VALUE)
        watcher_call_count = 0

        def watch_count(self, new_value: int) -> None:
            self.watcher_call_count += 1

    app = WatcherInitTrue()
    async with app.run_test():
        assert app.count == OLD_VALUE
        assert app.watcher_call_count == 1  # Watcher called on init
        app.count = NEW_VALUE  # User sets the value...
        assert app.watcher_call_count == 2  # ...resulting in 2nd call
        app.count = NEW_VALUE  # Setting to the SAME value
        assert app.watcher_call_count == 2  # Watcher is NOT called again


async def test_reactive_always_update():
    calls = []

    class AlwaysUpdate(App):
        first_name = reactive("Darren", init=False, always_update=True)
        last_name = reactive("Burns", init=False)

        def watch_first_name(self, value):
            calls.append(f"first_name {value}")

        def watch_last_name(self, value):
            calls.append(f"last_name {value}")

    app = AlwaysUpdate()
    async with app.run_test():
        # Value is the same, but always_update=True, so watcher called...
        app.first_name = "Darren"
        assert calls == ["first_name Darren"]
        # Value is the same, and always_update=False, so watcher NOT called...
        app.last_name = "Burns"
        assert calls == ["first_name Darren"]
        # Values changed, watch method always called regardless of always_update
        app.first_name = "abc"
        app.last_name = "def"
        assert calls == ["first_name Darren", "first_name abc", "last_name def"]


async def test_reactive_with_callable_default():
    """A callable can be supplied as the default value for a reactive.
    Textual will call it in order to retrieve the default value."""

    class ReactiveCallable(App):
        value = reactive(lambda: 123)
        watcher_called_with = None

        def watch_value(self, new_value):
            self.watcher_called_with = new_value

    app = ReactiveCallable()
    async with app.run_test():
        assert app.value == 123
        assert app.watcher_called_with == 123


async def test_validate_init_true():
    """When init is True for a reactive attribute, Textual should call the validator
    AND the watch method when the app starts."""
    validator_call_count = 0

    class ValidatorInitTrue(App):
        count = var(5, init=True)

        def validate_count(self, value: int) -> int:
            nonlocal validator_call_count
            validator_call_count += 1
            return value + 1

    app = ValidatorInitTrue()
    async with app.run_test():
        app.count = 5
        assert app.count == 6  # Validator should run, so value should be 5+1=6
        assert validator_call_count == 1


async def test_validate_init_true_set_before_dom_ready():
    """When init is True for a reactive attribute, Textual should call the validator
    AND the watch method when the app starts."""
    validator_call_count = 0

    class ValidatorInitTrue(App):
        count = var(5, init=True)

        def validate_count(self, value: int) -> int:
            nonlocal validator_call_count
            validator_call_count += 1
            return value + 1

    app = ValidatorInitTrue()
    app.count = 5
    async with app.run_test():
        assert app.count == 6  # Validator should run, so value should be 5+1=6
        assert validator_call_count == 1


async def test_reactive_compute_first_time_set():
    class ReactiveComputeFirstTimeSet(App):
        number = reactive(1)
        double_number = reactive(None)

        def compute_double_number(self):
            return self.number * 2

    app = ReactiveComputeFirstTimeSet()
    async with app.run_test():
        assert app.double_number == 2


async def test_reactive_method_call_order():
    class CallOrder(App):
        count = reactive(OLD_VALUE, init=False)
        count_times_ten = reactive(OLD_VALUE * 10, init=False)
        calls = []

        def validate_count(self, value: int) -> int:
            self.calls.append(f"validate {value}")
            return value + 1

        def watch_count(self, value: int) -> None:
            self.calls.append(f"watch {value}")

        def compute_count_times_ten(self) -> int:
            self.calls.append(f"compute {self.count}")
            return self.count * 10

    app = CallOrder()
    async with app.run_test():
        app.count = NEW_VALUE
        assert app.calls == [
            # The validator receives NEW_VALUE, since that's what the user
            # set the reactive attribute to...
            f"validate {NEW_VALUE}",
            # The validator adds 1 to the new value, and this is what should
            # be passed into the watcher...
            f"watch {NEW_VALUE + 1}",
            # The compute method accesses the reactive value directly, which
            # should have been updated by the validator to NEW_VALUE + 1.
            f"compute {NEW_VALUE + 1}",
        ]
        assert app.count == NEW_VALUE + 1
        assert app.count_times_ten == (NEW_VALUE + 1) * 10


async def test_premature_reactive_call():
    watcher_called = False

    class BrokenWidget(Widget):
        foo = reactive(1)

        def __init__(self) -> None:
            super().__init__()
            self.foo = "bar"

        async def watch_foo(self) -> None:
            nonlocal watcher_called
            watcher_called = True

    class PrematureApp(App):
        def compose(self) -> ComposeResult:
            yield BrokenWidget()

    app = PrematureApp()
    async with app.run_test() as pilot:
        assert watcher_called
        app.exit()


async def test_reactive_inheritance():
    """Check that inheritance works as expected for reactives."""

    class Primary(App):
        foo = reactive(1)
        bar = reactive("bar")

    class Secondary(Primary):
        foo = reactive(2)
        egg = reactive("egg")

    class Tertiary(Secondary):
        baz = reactive("baz")

    primary = Primary()
    secondary = Secondary()
    tertiary = Tertiary()

    primary_reactive_count = len(primary._reactives)

    # Secondary adds one new reactive
    assert len(secondary._reactives) == primary_reactive_count + 1

    Reactive._initialize_object(primary)
    Reactive._initialize_object(secondary)
    Reactive._initialize_object(tertiary)

    # Primary doesn't have egg
    with pytest.raises(AttributeError):
        assert primary.egg

    # primary has foo of 1
    assert primary.foo == 1
    # secondary has different reactive
    assert secondary.foo == 2
    # foo is accessible through tertiary
    assert tertiary.foo == 2

    with pytest.raises(AttributeError):
        secondary.baz

    assert tertiary.baz == "baz"


async def test_compute():
    """Check compute method is called."""

    class ComputeApp(App):
        count = var(0)
        count_double = var(0)

        def __init__(self) -> None:
            self.start = 0
            super().__init__()

        def compute_count_double(self) -> int:
            return self.start + self.count * 2

    app = ComputeApp()

    async with app.run_test():
        assert app.count_double == 0
        app.count = 1
        assert app.count_double == 2
        assert app.count_double == 2
        app.count = 2
        assert app.count_double == 4
        app.start = 10
        assert app.count_double == 14

        with pytest.raises(AttributeError):
            app.count_double = 100


async def test_watch_compute():
    """Check that watching a computed attribute works."""

    watch_called: list[bool] = []

    class Calculator(App):
        numbers = var("0")
        show_ac = var(True)
        value = var("")

        def compute_show_ac(self) -> bool:
            return self.value in ("", "0") and self.numbers == "0"

        def watch_show_ac(self, show_ac: bool) -> None:
            """Called when show_ac changes."""
            watch_called.append(show_ac)

    app = Calculator()

    # Referencing the value calls compute
    # Setting any reactive values calls compute
    async with app.run_test():
        assert app.show_ac is True
        app.value = "1"
        assert app.show_ac is False
        app.value = "0"
        assert app.show_ac is True
        app.numbers = "123"
        assert app.show_ac is False

    assert watch_called == [True, True, False, False, True, True, False, False]


async def test_public_and_private_watch() -> None:
    """If a reactive/var has public and private watches both should get called."""

    calls: dict[str, bool] = {"private": False, "public": False}

    class PrivateWatchTest(App):
        counter = var(0, init=False)

        def watch_counter(self) -> None:
            calls["public"] = True

        def _watch_counter(self) -> None:
            calls["private"] = True

    async with PrivateWatchTest().run_test() as pilot:
        assert calls["private"] is False
        assert calls["public"] is False
        pilot.app.counter += 1
        assert calls["private"] is True
        assert calls["public"] is True


async def test_private_validate() -> None:
    calls: dict[str, bool] = {"private": False}

    class PrivateValidateTest(App):
        counter = var(0, init=False)

        def _validate_counter(self, _: int) -> None:
            calls["private"] = True

    async with PrivateValidateTest().run_test() as pilot:
        assert calls["private"] is False
        pilot.app.counter += 1
        assert calls["private"] is True


async def test_public_and_private_validate() -> None:
    """If a reactive/var has public and private validate both should get called."""

    calls: dict[str, bool] = {"private": False, "public": False}

    class PrivateValidateTest(App):
        counter = var(0, init=False)

        def validate_counter(self, _: int) -> None:
            calls["public"] = True

        def _validate_counter(self, _: int) -> None:
            calls["private"] = True

    async with PrivateValidateTest().run_test() as pilot:
        assert calls["private"] is False
        assert calls["public"] is False
        pilot.app.counter += 1
        assert calls["private"] is True
        assert calls["public"] is True


async def test_public_and_private_validate_order() -> None:
    """The private validate should be called first."""

    class ValidateOrderTest(App):
        value = var(0, init=False)

        def validate_value(self, value: int) -> int:
            if value < 0:
                return 42
            return value

        def _validate_value(self, value: int) -> int:
            if value < 0:
                return 73
            return value

    async with ValidateOrderTest().run_test() as pilot:
        pilot.app.value = -10
        assert pilot.app.value == 73


async def test_public_and_private_compute() -> None:
    """If a reactive/var has public and private compute both should get called."""

    with pytest.raises(TooManyComputesError):

        class PublicAndPrivateComputeTest(App):
            counter = var(0, init=False)

            def compute_counter(self):
                pass

            def _compute_counter(self):
                pass


async def test_private_compute() -> None:
    class PrivateComputeTest(App):
        double = var(0, init=False)
        base = var(0, init=False)

        def _compute_double(self) -> int:
            return 2 * self.base

    async with PrivateComputeTest().run_test() as pilot:
        pilot.app.base = 5
        assert pilot.app.double == 10


async def test_async_reactive_watch_callbacks_go_on_the_watcher():
    """Regression test for https://github.com/Textualize/textual/issues/3036.

    This makes sure that async callbacks are called.
    See the next test for sync callbacks.
    """

    from_app = False
    from_holder = False

    class Holder(Widget):
        attr = var(None)

        def watch_attr(self):
            nonlocal from_holder
            from_holder = True

    class MyApp(App):
        def __init__(self):
            super().__init__()
            self.holder = Holder()

        def on_mount(self):
            self.watch(self.holder, "attr", self.callback)

        def update(self):
            self.holder.attr = "hello world"

        async def callback(self):
            nonlocal from_app
            from_app = True

    async with MyApp().run_test() as pilot:
        pilot.app.update()
        await pilot.pause()
        assert from_holder
        assert from_app


async def test_sync_reactive_watch_callbacks_go_on_the_watcher():
    """Regression test for https://github.com/Textualize/textual/issues/3036.

    This makes sure that sync callbacks are called.
    See the previous test for async callbacks.
    """

    from_app = False
    from_holder = False

    class Holder(Widget):
        attr = var(None)

        def watch_attr(self):
            nonlocal from_holder
            from_holder = True

    class MyApp(App):
        def __init__(self):
            super().__init__()
            self.holder = Holder()

        def on_mount(self):
            self.watch(self.holder, "attr", self.callback)

        def update(self):
            self.holder.attr = "hello world"

        def callback(self):
            nonlocal from_app
            from_app = True

    async with MyApp().run_test() as pilot:
        pilot.app.update()
        await pilot.pause()
        assert from_holder
        assert from_app


async def test_set_reactive():
    """Test set_reactive doesn't call watchers."""

    class MyWidget(Widget):
        foo = reactive("")

        def __init__(self, foo: str) -> None:
            super().__init__()
            self.set_reactive(MyWidget.foo, foo)

        def watch_foo(self) -> None:
            # Should never get here
            1 / 0

    class MyApp(App):
        def compose(self) -> ComposeResult:
            yield MyWidget("foobar")

    app = MyApp()
    async with app.run_test():
        assert app.query_one(MyWidget).foo == "foobar"


async def test_no_duplicate_external_watchers() -> None:
    """Make sure we skip duplicated watchers."""

    counter = 0

    class Holder(Widget):
        attr = var(None)

    class MyApp(App[None]):
        def __init__(self) -> None:
            super().__init__()
            self.holder = Holder()

        def on_mount(self) -> None:
            self.watch(self.holder, "attr", self.callback)
            self.watch(self.holder, "attr", self.callback)

        def callback(self) -> None:
            nonlocal counter
            counter += 1

    app = MyApp()
    async with app.run_test():
        assert counter == 1
        app.holder.attr = 73
        assert counter == 2


async def test_external_watch_init_does_not_propagate() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3878.

    Make sure that when setting an extra watcher programmatically and `init` is set,
    we init only the new watcher and not the other ones, but at the same
    time make sure both watchers work in regular circumstances.
    """

    logs: list[str] = []

    class SomeWidget(Widget):
        test_1: var[int] = var(0)
        test_2: var[int] = var(0, init=False)

        def watch_test_1(self) -> None:
            logs.append("test_1")

        def watch_test_2(self) -> None:
            logs.append("test_2")

    class InitOverrideApp(App[None]):
        def compose(self) -> ComposeResult:
            yield SomeWidget()

        def on_mount(self) -> None:
            def watch_test_2_extra() -> None:
                logs.append("test_2_extra")

            self.watch(self.query_one(SomeWidget), "test_2", watch_test_2_extra)

    app = InitOverrideApp()
    async with app.run_test():
        assert logs == ["test_1", "test_2_extra"]
        app.query_one(SomeWidget).test_2 = 73
        assert logs.count("test_2_extra") == 2
        assert logs.count("test_2") == 1


async def test_external_watch_init_does_not_propagate_to_externals() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3878.

    Make sure that when setting an extra watcher programmatically and `init` is set,
    we init only the new watcher and not the other ones (even if they were
    added dynamically with `watch`), but at the same time make sure all watchers
    work in regular circumstances.
    """

    logs: list[str] = []

    class SomeWidget(Widget):
        test_var: var[int] = var(0)

    class MyApp(App[None]):
        def compose(self) -> ComposeResult:
            yield SomeWidget()

        def add_first_watcher(self) -> None:
            def first_callback() -> None:
                logs.append("first")

            self.watch(self.query_one(SomeWidget), "test_var", first_callback)

        def add_second_watcher(self) -> None:
            def second_callback() -> None:
                logs.append("second")

            self.watch(self.query_one(SomeWidget), "test_var", second_callback)

    app = MyApp()
    async with app.run_test():
        assert logs == []
        app.add_first_watcher()
        assert logs == ["first"]
        app.add_second_watcher()
        assert logs == ["first", "second"]
        app.query_one(SomeWidget).test_var = 73
        assert logs == ["first", "second", "first", "second"]


async def test_message_sender_from_reactive() -> None:
    """Test that the sender of a message comes from the reacting widget."""

    message_senders: list[MessagePump | None] = []

    class TestWidget(Widget):
        test_var: var[int] = var(0, init=False)

        class TestMessage(Message):
            pass

        def watch_test_var(self) -> None:
            self.post_message(self.TestMessage())

        def make_reaction(self) -> None:
            self.test_var += 1

    class TestContainer(Widget):
        def compose(self) -> ComposeResult:
            yield TestWidget()

        def on_test_widget_test_message(self, event: TestWidget.TestMessage) -> None:
            nonlocal message_senders
            message_senders.append(event._sender)

    class TestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield TestContainer()

    async with TestApp().run_test() as pilot:
        assert message_senders == []
        pilot.app.query_one(TestWidget).make_reaction()
        await pilot.pause()
        assert message_senders == [pilot.app.query_one(TestWidget)]


async def test_mutate_reactive() -> None:
    """Test explicitly mutating reactives"""

    watched_names: list[list[str]] = []

    class TestWidget(Widget):
        names: reactive[list[str]] = reactive(list)

        def watch_names(self, names: list[str]) -> None:
            watched_names.append(names.copy())

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield TestWidget()

    app = TestApp()
    async with app.run_test():
        widget = app.query_one(TestWidget)
        # watch method called on startup
        assert watched_names == [[]]

        # Mutate the list
        widget.names.append("Paul")
        # No changes expected
        assert watched_names == [[]]
        # Explicitly mutate the reactive
        widget.mutate_reactive(TestWidget.names)
        # Watcher will be invoked
        assert watched_names == [[], ["Paul"]]
        # Make further modifications
        widget.names.append("Jessica")
        widget.names.remove("Paul")
        # No change expected
        assert watched_names == [[], ["Paul"]]
        # Explicit mutation
        widget.mutate_reactive(TestWidget.names)
        # Watcher should be invoked
        assert watched_names == [[], ["Paul"], ["Jessica"]]


async def test_mutate_reactive_data_bind() -> None:
    """https://github.com/Textualize/textual/issues/4825"""

    # Record mutations to TestWidget.messages
    widget_messages: list[list[str]] = []

    class TestWidget(Widget):
        messages: reactive[list[str]] = reactive(list, init=False)

        def watch_messages(self, names: list[str]) -> None:
            widget_messages.append(names.copy())

    class TestApp(App):
        messages: reactive[list[str]] = reactive(list, init=False)

        def compose(self) -> ComposeResult:
            yield TestWidget().data_bind(TestApp.messages)

    app = TestApp()
    async with app.run_test():
        test_widget = app.query_one(TestWidget)
        assert widget_messages == [[]]
        assert test_widget.messages == []

        # Should be the same instance
        assert app.messages is test_widget.messages

        # Mutate app
        app.messages.append("foo")
        # Mutations aren't detected
        assert widget_messages == [[]]
        assert app.messages == ["foo"]
        assert test_widget.messages == ["foo"]
        # Explicitly mutate app reactive
        app.mutate_reactive(TestApp.messages)
        # Mutating app, will also invoke watchers on any data binds
        assert widget_messages == [[], ["foo"]]
        assert app.messages == ["foo"]
        assert test_widget.messages == ["foo"]



from textual.app import App
from textual.containers import Horizontal
from textual.widgets import Placeholder


class MinHeightApp(App):
    CSS_PATH = "min_height.tcss"

    def compose(self):
        yield Horizontal(
            Placeholder("min-height: 25%", id="p1"),
            Placeholder("min-height: 75%", id="p2"),
            Placeholder("min-height: 30", id="p3"),
            Placeholder("min-height: 40w", id="p4"),
        )


if __name__ == "__main__":
    app = MinHeightApp()
    app.run()



from __future__ import annotations

import ast
import re
from functools import lru_cache
from typing import Any

from typing_extensions import TypeAlias

ActionParseResult: TypeAlias = "tuple[str, str, tuple[object, ...]]"
"""An action is its name and the arbitrary tuple of its arguments."""


class SkipAction(Exception):
    """Raise in an action to skip the action (and allow any parent bindings to run)."""


class ActionError(Exception):
    pass


re_action_args = re.compile(r"([\w\.]+)\((.*)\)")


@lru_cache(maxsize=1024)
def parse(action: str) -> ActionParseResult:
    """Parses an action string.

    Args:
        action: String containing action.

    Raises:
        ActionError: If the action has invalid syntax.

    Returns:
        Action name and arguments.
    """
    args_match = re_action_args.match(action)
    if args_match is not None:
        action_name, action_args_str = args_match.groups()
        if action_args_str:
            try:
                # We wrap `action_args_str` to be able to disambiguate the cases where
                # the list of arguments is a comma-separated list of values from the
                # case where the argument is a single tuple.
                action_args: tuple[Any, ...] = ast.literal_eval(f"({action_args_str},)")
            except Exception:
                raise ActionError(
                    f"unable to parse {action_args_str!r} in action {action!r}"
                )
        else:
            action_args = ()
    else:
        action_name = action
        action_args = ()

    namespace, _, action_name = action_name.rpartition(".")

    return namespace, action_name, action_args



from rich.syntax import Syntax

from textual.app import App, ComposeResult, RenderResult
from textual.reactive import reactive
from textual.widget import Widget


class CodeView(Widget):
    """Widget to display Python code."""

    DEFAULT_CSS = """
    CodeView { height: auto; }
    """

    code = reactive("")

    def render(self) -> RenderResult:
        # Syntax is a Rich renderable that displays syntax highlighted code
        syntax = Syntax(self.code, "python", line_numbers=True, indent_guides=True)
        return syntax


class CodeApp(App):
    """App to demonstrate Rich renderables in Textual."""

    def compose(self) -> ComposeResult:
        with open(__file__) as self_file:
            code = self_file.read()
        code_view = CodeView()
        code_view.code = code
        yield code_view


if __name__ == "__main__":
    app = CodeApp()
    app.run()



import pytest

from textual._duration import DurationParseError, _duration_as_seconds


def test_parse() -> None:
    assert _duration_as_seconds("30") == 30.0
    assert _duration_as_seconds("30s") == 30.0
    assert _duration_as_seconds("30000ms") == 30.0
    assert _duration_as_seconds("0.5") == 0.5
    assert _duration_as_seconds("0.5s") == 0.5
    assert _duration_as_seconds("500ms") == 0.5

    with pytest.raises(DurationParseError):
        _duration_as_seconds("300x")



from textual.app import App, ComposeResult
from textual.widgets import Static


class PrideApp(App):
    """Displays a pride flag."""

    COLORS = ["red", "orange", "yellow", "green", "blue", "purple"]

    def compose(self) -> ComposeResult:
        for color in self.COLORS:
            stripe = Static()
            stripe.styles.height = "1fr"
            stripe.styles.background = color
            yield stripe


if __name__ == "__main__":
    PrideApp().run()



from textual.app import App
from textual.containers import Horizontal
from textual.widgets import Footer, Tree


class TreeApp(App[None]):
    def compose(self):
        yield Horizontal(Tree("Dune"))
        yield Footer()


async def test_shutdown():
    # regression test for https://github.com/Textualize/textual/issues/4634
    # Testing that an app with the footer doesn't deadlock
    app = TreeApp()
    print("Check for deadlock in test_shutdown.py")
    async with app.run_test():
        pass



"""
Define a series of easing functions for more natural-looking animations.
Taken from https://easings.net/ and translated from JavaScript.
"""

from math import cos, pi, sin, sqrt


def _in_out_expo(x: float) -> float:
    """https://easings.net/#easeInOutExpo"""
    if 0 < x < 0.5:
        return pow(2, 20 * x - 10) / 2
    elif 0.5 <= x < 1:
        return (2 - pow(2, -20 * x + 10)) / 2
    else:
        return x  # x in (0, 1)


def _in_out_circ(x: float) -> float:
    """https://easings.net/#easeInOutCirc"""
    if x < 0.5:
        return (1 - sqrt(1 - pow(2 * x, 2))) / 2
    else:
        return (sqrt(1 - pow(-2 * x + 2, 2)) + 1) / 2


def _in_out_back(x: float) -> float:
    """https://easings.net/#easeInOutBack"""
    c = 1.70158 * 1.525
    if x < 0.5:
        return (pow(2 * x, 2) * ((c + 1) * 2 * x - c)) / 2
    else:
        return (pow(2 * x - 2, 2) * ((c + 1) * (x * 2 - 2) + c) + 2) / 2


def _in_elastic(x: float) -> float:
    """https://easings.net/#easeInElastic"""
    c = 2 * pi / 3
    if 0 < x < 1:
        return -pow(2, 10 * x - 10) * sin((x * 10 - 10.75) * c)
    else:
        return x  # x in (0, 1)


def _in_out_elastic(x: float) -> float:
    """https://easings.net/#easeInOutElastic"""
    c = 2 * pi / 4.5
    if 0 < x < 0.5:
        return -(pow(2, 20 * x - 10) * sin((20 * x - 11.125) * c)) / 2
    elif 0.5 <= x < 1:
        return (pow(2, -20 * x + 10) * sin((20 * x - 11.125) * c)) / 2 + 1
    else:
        return x  # x in (0, 1)


def _out_elastic(x: float) -> float:
    """https://easings.net/#easeInOutElastic"""
    c = 2 * pi / 3
    if 0 < x < 1:
        return pow(2, -10 * x) * sin((x * 10 - 0.75) * c) + 1
    else:
        return x  # x in (0, 1)


def _out_bounce(x: float) -> float:
    """https://easings.net/#easeOutBounce"""
    n, d = 7.5625, 2.75
    if x < 1 / d:
        return n * x * x
    elif x < 2 / d:
        x_ = x - 1.5 / d
        return n * x_ * x_ + 0.75
    elif x < 2.5 / d:
        x_ = x - 2.25 / d
        return n * x_ * x_ + 0.9375
    else:
        x_ = x - 2.625 / d
        return n * x_ * x_ + 0.984375


def _in_bounce(x: float) -> float:
    """https://easings.net/#easeInBounce"""
    return 1 - _out_bounce(1 - x)


def _in_out_bounce(x: float) -> float:
    """https://easings.net/#easeInOutBounce"""
    if x < 0.5:
        return (1 - _out_bounce(1 - 2 * x)) / 2
    else:
        return (1 + _out_bounce(2 * x - 1)) / 2


EASING = {
    "none": lambda x: 1.0,
    "round": lambda x: 0.0 if x < 0.5 else 1.0,
    "linear": lambda x: x,
    "in_sine": lambda x: 1 - cos((x * pi) / 2),
    "in_out_sine": lambda x: -(cos(x * pi) - 1) / 2,
    "out_sine": lambda x: sin((x * pi) / 2),
    "in_quad": lambda x: x * x,
    "in_out_quad": lambda x: 2 * x * x if x < 0.5 else 1 - pow(-2 * x + 2, 2) / 2,
    "out_quad": lambda x: 1 - pow(1 - x, 2),
    "in_cubic": lambda x: x * x * x,
    "in_out_cubic": lambda x: 4 * x * x * x if x < 0.5 else 1 - pow(-2 * x + 2, 3) / 2,
    "out_cubic": lambda x: 1 - pow(1 - x, 3),
    "in_quart": lambda x: pow(x, 4),
    "in_out_quart": lambda x: 8 * pow(x, 4) if x < 0.5 else 1 - pow(-2 * x + 2, 4) / 2,
    "out_quart": lambda x: 1 - pow(1 - x, 4),
    "in_quint": lambda x: pow(x, 5),
    "in_out_quint": lambda x: 16 * pow(x, 5) if x < 0.5 else 1 - pow(-2 * x + 2, 5) / 2,
    "out_quint": lambda x: 1 - pow(1 - x, 5),
    "in_expo": lambda x: pow(2, 10 * x - 10) if x else 0,
    "in_out_expo": _in_out_expo,
    "out_expo": lambda x: 1 - pow(2, -10 * x) if x != 1 else 1,
    "in_circ": lambda x: 1 - sqrt(1 - pow(x, 2)),
    "in_out_circ": _in_out_circ,
    "out_circ": lambda x: sqrt(1 - pow(x - 1, 2)),
    "in_back": lambda x: 2.70158 * pow(x, 3) - 1.70158 * pow(x, 2),
    "in_out_back": _in_out_back,
    "out_back": lambda x: 1 + 2.70158 * pow(x - 1, 3) + 1.70158 * pow(x - 1, 2),
    "in_elastic": _in_elastic,
    "in_out_elastic": _in_out_elastic,
    "out_elastic": _out_elastic,
    "in_bounce": _in_bounce,
    "in_out_bounce": _in_out_bounce,
    "out_bounce": _out_bounce,
}

DEFAULT_EASING = "in_out_cubic"
DEFAULT_SCROLL_EASING = "out_cubic"



from textual.app import App, ComposeResult
from textual.widgets import Label


class ContentApp(App):

    def compose(self) -> ComposeResult:
        yield Label("[bold]Bold[/] [italic]Italic[/] [u]Underline[/]  [s]Strike[/s]")
        yield Label(
            "[$primary]Primary[/]  [$secondary]Secondary[/]  [$warning]Warning[/]  [$error]Error[/]"
        )
        yield Label("[$text on $primary]Text on Primary")
        yield Label("[$primary on $primary-muted]Primary on primary muted")
        yield Label("[$error on $error-muted]Error on error muted")
        yield Label(
            "[on $boost]    [on $boost]    [on $boost]  Three layers of $boost  [/]    [/]    [/]"
        )
        yield Label("[on $primary 20%]On primary twenty percent")
        yield Label("[$text 80% on $primary]Hello")


if __name__ == "__main__":
    app = ContentApp()
    app.run()



import pytest

from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget


@pytest.fixture
async def app():
    class HorizontalAutoWidth(App):
        def compose(self) -> ComposeResult:
            child1 = Widget(id="child1")
            child1.styles.width = 4
            child2 = Widget(id="child2")
            child2.styles.width = 6
            child3 = Widget(id="child3")
            child3.styles.width = 5
            self.horizontal = Horizontal(child1, child2, child3)
            yield self.horizontal

    app = HorizontalAutoWidth()
    yield app


async def test_horizontal_get_content_width(app):
    async with app.run_test():
        size = app.screen.size
        width = app.horizontal.get_content_width(size, size)
        assert width == 15



from __future__ import annotations

try:
    import httpx
except ImportError:
    raise ImportError("Please install httpx with 'pip install httpx' ")


from textual import work
from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Input, Markdown


class DictionaryApp(App):
    """Searches a dictionary API as-you-type."""

    CSS_PATH = "dictionary.tcss"

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Search for a word", id="dictionary-search")
        with VerticalScroll(id="results-container"):
            yield Markdown(id="results")

    async def on_input_changed(self, message: Input.Changed) -> None:
        """A coroutine to handle a text changed message."""
        if message.value:
            self.lookup_word(message.value)
        else:
            # Clear the results
            await self.query_one("#results", Markdown).update("")

    @work(exclusive=True)
    async def lookup_word(self, word: str) -> None:
        """Looks up a word."""
        url = f"https://api.dictionaryapi.dev/api/v2/entries/en/{word}"

        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            try:
                results = response.json()
            except Exception:
                self.query_one("#results", Markdown).update(response.text)
                return

        if word == self.query_one(Input).value:
            markdown = self.make_word_markdown(results)
            self.query_one("#results", Markdown).update(markdown)

    def make_word_markdown(self, results: object) -> str:
        """Convert the results into markdown."""
        lines = []
        if isinstance(results, dict):
            lines.append(f"# {results['title']}")
            lines.append(results["message"])
        elif isinstance(results, list):
            for result in results:
                lines.append(f"# {result['word']}")
                lines.append("")
                for meaning in result.get("meanings", []):
                    lines.append(f"_{meaning['partOfSpeech']}_")
                    lines.append("")
                    for definition in meaning.get("definitions", []):
                        lines.append(f" - {definition['definition']}")
                    lines.append("---")

        return "\n".join(lines)


if __name__ == "__main__":
    app = DictionaryApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Grid
from textual.widgets import Placeholder


class KeylineApp(App):
    CSS_PATH = "keyline.tcss"

    def compose(self) -> ComposeResult:
        with Grid():
            yield Placeholder(id="foo")
            yield Placeholder(id="bar")
            yield Placeholder()
            yield Placeholder(classes="hidden")
            yield Placeholder(id="baz")


if __name__ == "__main__":
    KeylineApp().run()



import json
from pathlib import Path

from rich.text import Text

from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, Tree
from textual.widgets.tree import TreeNode


class TreeApp(App):

    BINDINGS = [
        ("a", "add", "Add node"),
        ("c", "clear", "Clear"),
        ("t", "toggle_root", "Toggle root"),
    ]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Tree("Root")

    @classmethod
    def add_json(cls, node: TreeNode, json_data: object) -> None:
        """Adds JSON data to a node.

        Args:
            node (TreeNode): A Tree node.
            json_data (object): An object decoded from JSON.
        """

        from rich.highlighter import ReprHighlighter

        highlighter = ReprHighlighter()

        def add_node(name: str, node: TreeNode, data: object) -> None:
            """Adds a node to the tree.

            Args:
                name (str): Name of the node.
                node (TreeNode): Parent node.
                data (object): Data associated with the node.
            """
            if isinstance(data, dict):
                node.set_label(Text(f"{{}} {name}"))
                for key, value in data.items():
                    new_node = node.add("")
                    add_node(key, new_node, value)
            elif isinstance(data, list):
                node.set_label(Text(f"[] {name}"))
                for index, value in enumerate(data):
                    new_node = node.add("")
                    add_node(str(index), new_node, value)
            else:
                node.allow_expand = False
                if name:
                    label = Text.assemble(
                        Text.from_markup(f"[b]{name}[/b]="), highlighter(repr(data))
                    )
                else:
                    label = Text(repr(data))
                node.set_label(label)

        add_node("JSON", node, json_data)

    def on_mount(self) -> None:
        """Load some JSON when the app starts."""
        file_path = Path(__file__).parent / "food.json"
        with open(file_path) as data_file:
            self.json_data = json.load(data_file)

    def action_add(self) -> None:
        """Add a node to the tree."""
        tree = self.query_one(Tree)
        json_node = tree.root.add("JSON")
        self.add_json(json_node, self.json_data)
        tree.root.expand()

    def action_clear(self) -> None:
        """Clear the tree (remove all nodes)."""
        tree = self.query_one(Tree)
        tree.clear()

    def action_toggle_root(self) -> None:
        """Toggle the root node."""
        tree = self.query_one(Tree)
        tree.show_root = not tree.show_root


if __name__ == "__main__":
    app = TreeApp()
    app.run()



from textual.app import App
from textual.containers import Grid
from textual.widgets import Label


class MyApp(App):
    CSS_PATH = "grid_rows.tcss"

    def compose(self):
        yield Grid(
            Label("1fr"),
            Label("1fr"),
            Label("height = 6"),
            Label("height = 6"),
            Label("25%"),
            Label("25%"),
            Label("1fr"),
            Label("1fr"),
            Label("height = 6"),
            Label("height = 6"),
        )


if __name__ == "__main__":
    app = MyApp()
    app.run()



from textual.app import App
from textual import events


class ActionsApp(App):
    def action_set_background(self, color: str) -> None:
        self.screen.styles.background = color

    def on_key(self, event: events.Key) -> None:
        if event.key == "r":
            self.action_set_background("red")


if __name__ == "__main__":
    app = ActionsApp()
    app.run()



from textual.app import App, ComposeResult
from textual.color import Color
from textual.widgets import Static


class ColorApp(App):
    def compose(self) -> ComposeResult:
        self.widgets = [Static("") for n in range(10)]
        yield from self.widgets

    def on_mount(self) -> None:
        for index, widget in enumerate(self.widgets, 1):
            alpha = index * 0.1
            widget.update(f"alpha={alpha:.1f}")
            widget.styles.background = Color(191, 78, 96, a=alpha)


if __name__ == "__main__":
    app = ColorApp()
    app.run()



import pytest

from textual.app import App
from textual.widgets import Select
from textual.widgets.select import InvalidSelectValueError

SELECT_OPTIONS = [(str(n), n) for n in range(3)]
MORE_OPTIONS = [(str(n), n) for n in range(5, 8)]


class SelectApp(App[None]):
    def __init__(self, initial_value=Select.BLANK):
        self.initial_value = initial_value
        super().__init__()

    def compose(self):
        yield Select[int](SELECT_OPTIONS, value=self.initial_value)


async def test_initial_value_is_validated():
    """The initial value should be respected if it is a legal value.

    Regression test for https://github.com/Textualize/textual/discussions/3037.
    """
    app = SelectApp(1)
    async with app.run_test():
        assert app.query_one(Select).value == 1


async def test_value_unknown_option_raises_error():
    """Setting the value to an unknown value raises an error."""
    app = SelectApp()
    async with app.run_test():
        with pytest.raises(InvalidSelectValueError):
            app.query_one(Select).value = "french fries"


async def test_initial_value_inside_compose_is_validated():
    """Setting the value to an unknown value inside compose should raise an error."""

    class SelectApp(App[None]):
        def compose(self):
            s = Select[int](SELECT_OPTIONS)
            s.value = 73
            yield s

    app = SelectApp()
    with pytest.raises(InvalidSelectValueError):
        async with app.run_test():
            pass


async def test_value_assign_to_blank():
    """Setting the value to BLANK should work with default `allow_blank` value."""
    app = SelectApp(1)
    async with app.run_test():
        select = app.query_one(Select)
        assert select.value == 1
        select.value = Select.BLANK
        assert select.is_blank()


async def test_initial_value_is_picked_if_allow_blank_is_false():
    """The initial value should be picked by default if allow_blank=False."""

    class SelectApp(App[None]):
        def compose(self):
            yield Select[int](SELECT_OPTIONS, allow_blank=False)

    app = SelectApp()
    async with app.run_test():
        assert app.query_one(Select).value == 0


async def test_initial_value_is_picked_if_allow_blank_is_false():
    """The initial value should be respected even if allow_blank=False."""

    class SelectApp(App[None]):
        def compose(self):
            yield Select[int](SELECT_OPTIONS, value=2, allow_blank=False)

    app = SelectApp()
    async with app.run_test():
        assert app.query_one(Select).value == 2


async def test_set_value_to_blank_with_allow_blank_false():
    """Setting the value to BLANK with allow_blank=False should raise an error."""

    class SelectApp(App[None]):
        def compose(self):
            yield Select[int](SELECT_OPTIONS, allow_blank=False)

    app = SelectApp()
    async with app.run_test():
        with pytest.raises(InvalidSelectValueError):
            app.query_one(Select).value = Select.BLANK


async def test_set_options_resets_value_to_blank():
    """Resetting the options should reset the value to BLANK."""

    class SelectApp(App[None]):
        def compose(self):
            yield Select[int](SELECT_OPTIONS, value=2)

    app = SelectApp()
    async with app.run_test():
        select = app.query_one(Select)
        assert select.value == 2
        select.set_options(MORE_OPTIONS)
        assert select.is_blank()


async def test_set_options_resets_value_if_allow_blank_is_false():
    """Resetting the options should reset the value if allow_blank=False."""

    class SelectApp(App[None]):
        def compose(self):
            yield Select[int](SELECT_OPTIONS, allow_blank=False)

    app = SelectApp()
    async with app.run_test():
        select = app.query_one(Select)
        assert select.value == 0
        select.set_options(MORE_OPTIONS)
        assert select.value > 2



import itertools

import pytest

from textual._xterm_parser import XTermParser
from textual.events import (
    Key,
    MouseDown,
    MouseMove,
    MouseScrollDown,
    MouseScrollUp,
    MouseUp,
    Paste,
)
from textual.messages import TerminalSupportsSynchronizedOutput


def chunks(data, size):
    if size == 0:
        yield data
        return

    chunk_start = 0
    chunk_end = size
    while True:
        yield data[chunk_start:chunk_end]
        chunk_start = chunk_end
        chunk_end += size
        if chunk_end >= len(data):
            yield data[chunk_start:chunk_end]
            break


@pytest.fixture
def parser():
    return XTermParser()


@pytest.mark.parametrize("chunk_size", [2, 3, 4, 5, 6])
def test_varying_parser_chunk_sizes_no_missing_data(parser, chunk_size):
    end = "\x1b[8~"
    text = "ABCDEFGH"

    data = end + text
    events = []
    for chunk in chunks(data, chunk_size):
        events.append(parser.feed(chunk))

    events = list(itertools.chain.from_iterable(list(event) for event in events))

    assert events[0].key == "end"
    assert [event.key for event in events[1:]] == list(text)


def test_bracketed_paste(parser):
    """When bracketed paste mode is enabled in the terminal emulator and
    the user pastes in some text, it will surround the pasted input
    with the escape codes "\x1b[200~" and "\x1b[201~". The text between
    these codes corresponds to a single `Paste` event in Textual.
    """
    pasted_text = "PASTED"
    events = list(parser.feed(f"\x1b[200~{pasted_text}\x1b[201~"))

    assert len(events) == 1
    assert isinstance(events[0], Paste)
    assert events[0].text == pasted_text


def test_bracketed_paste_content_contains_escape_codes(parser):
    """When performing a bracketed paste, if the pasted content contains
    supported ANSI escape sequences, it should not interfere with the paste,
    and no escape sequences within the bracketed paste should be converted
    into Textual events.
    """
    pasted_text = "PAS\x0fTED"
    events = list(parser.feed(f"\x1b[200~{pasted_text}\x1b[201~"))
    assert len(events) == 1
    assert events[0].text == pasted_text


def test_bracketed_paste_amongst_other_codes(parser):
    pasted_text = "PASTED"
    events = list(parser.feed(f"\x1b[8~\x1b[200~{pasted_text}\x1b[201~\x1b[8~"))
    assert len(events) == 3  # Key.End -> Paste -> Key.End
    assert events[0].key == "end"
    assert events[1].text == pasted_text
    assert events[2].key == "end"


def test_cant_match_escape_sequence_too_long(parser):
    """The sequence did not match, and we hit the maximum sequence search
    length threshold, so each character should be issued as a key-press instead.
    """
    sequence = "\x1b[123456789123456789123123456789123456789123"
    events = list(parser.feed(sequence))

    # Every character in the sequence is converted to a key press
    assert len(events) == len(sequence)
    assert all(isinstance(event, Key) for event in events)

    # When we backtrack '\x1b' is translated to '^'
    assert events[0].key == "circumflex_accent"

    # The rest of the characters correspond to the expected key presses
    events = events[1:]
    for index, character in enumerate(sequence[1:]):
        assert events[index].character == character


@pytest.mark.parametrize(
    "chunk_size",
    [
        2,
        3,
        4,
        5,
        6,
    ],
)
def test_unknown_sequence_followed_by_known_sequence(parser, chunk_size):
    """When we feed the parser an unknown sequence followed by a known
    sequence. The characters in the unknown sequence are delivered as keys,
    and the known escape sequence that follows is delivered as expected.
    """
    unknown_sequence = "\x1b[?"
    known_sequence = "\x1b[8~"  # key = 'end'

    sequence = unknown_sequence + known_sequence

    events = []

    for chunk in chunks(sequence, chunk_size):
        events.append(parser.feed(chunk))

    events = list(itertools.chain.from_iterable(list(event) for event in events))
    print(repr([event.key for event in events]))

    assert [event.key for event in events] == [
        "escape",
        "left_square_bracket",
        "question_mark",
        "end",
    ]


def test_simple_key_presses_all_delivered_correct_order(parser):
    sequence = "123abc"
    events = parser.feed(sequence)
    assert "".join(event.key for event in events) == sequence


def test_simple_keypress_non_character_key(parser):
    sequence = "\x09"
    events = list(parser.feed(sequence))
    assert len(events) == 1
    assert events[0].key == "tab"


def test_key_presses_and_escape_sequence_mixed(parser):
    sequence = "abc\x1b[13~123"
    events = list(parser.feed(sequence))

    assert len(events) == 7
    assert "".join(event.key for event in events) == "abcf3123"


def test_single_escape(parser):
    """A single \x1b should be interpreted as a single press of the Escape key"""
    events = list(parser.feed("\x1b"))
    events.extend(parser.feed(""))
    assert [event.key for event in events] == ["escape"]


def test_double_escape(parser):
    """Test double escape."""
    events = list(parser.feed("\x1b\x1b"))
    events.extend(parser.feed(""))
    print(events)
    assert [event.key for event in events] == ["escape", "escape"]


@pytest.mark.parametrize(
    "sequence, event_type, shift, meta",
    [
        # Mouse down, with and without modifiers
        ("\x1b[<0;50;25M", MouseDown, False, False),
        ("\x1b[<4;50;25M", MouseDown, True, False),
        ("\x1b[<8;50;25M", MouseDown, False, True),
        ("\x1b[<12;50;25M", MouseDown, True, True),
        # Mouse up, with and without modifiers
        ("\x1b[<0;50;25m", MouseUp, False, False),
        ("\x1b[<4;50;25m", MouseUp, True, False),
        ("\x1b[<8;50;25m", MouseUp, False, True),
        ("\x1b[<12;50;25m", MouseUp, True, True),
    ],
)
def test_mouse_click(parser, sequence, event_type, shift, meta):
    """ANSI codes for mouse should be converted to Textual events"""
    events = list(parser.feed(sequence))

    assert len(events) == 1

    event = events[0]

    assert isinstance(event, event_type)
    assert event.x == 49
    assert event.y == 24
    assert event.screen_x == 49
    assert event.screen_y == 24
    assert event.meta is meta
    assert event.shift is shift


@pytest.mark.parametrize(
    "sequence, shift, meta, button",
    [
        ("\x1b[<32;15;38M", False, False, 1),  # Click and drag
        ("\x1b[<35;15;38M", False, False, 0),  # Basic cursor movement
        ("\x1b[<39;15;38M", True, False, 0),  # Shift held down
        ("\x1b[<43;15;38M", False, True, 0),  # Meta held down
        ("\x1b[<3;15;38M", False, False, 0),
    ],
)
def test_mouse_move(parser, sequence, shift, meta, button):
    events = list(parser.feed(sequence))

    assert len(events) == 1

    event = events[0]

    assert isinstance(event, MouseMove)
    assert event.x == 14
    assert event.y == 37
    assert event.shift is shift
    assert event.meta is meta
    assert event.button == button


@pytest.mark.parametrize(
    "sequence, shift, meta",
    [
        ("\x1b[<64;18;25M", False, False),
        ("\x1b[<68;18;25M", True, False),
        ("\x1b[<72;18;25M", False, True),
    ],
)
def test_mouse_scroll_up(parser, sequence, shift, meta):
    """Scrolling the mouse with and without modifiers held down.
    We don't currently capture modifier keys in scroll events.
    """
    events = list(parser.feed(sequence))

    assert len(events) == 1

    event = events[0]

    assert isinstance(event, MouseScrollUp)
    assert event.x == 17
    assert event.y == 24
    assert event.shift is shift
    assert event.meta is meta


@pytest.mark.parametrize(
    "sequence, shift, meta",
    [
        ("\x1b[<65;18;25M", False, False),
        ("\x1b[<69;18;25M", True, False),
        ("\x1b[<73;18;25M", False, True),
    ],
)
def test_mouse_scroll_down(parser, sequence, shift, meta):
    events = list(parser.feed(sequence))

    assert len(events) == 1

    event = events[0]

    assert isinstance(event, MouseScrollDown)
    assert event.x == 17
    assert event.y == 24
    assert event.shift is shift
    assert event.meta is meta


def test_mouse_event_detected_but_info_not_parsed(parser):
    # I don't know if this can actually happen in reality, but
    # there's a branch in the code that allows for the possibility.
    events = list(parser.feed("\x1b[<65;18;20;25M"))
    assert len(events) == 0


@pytest.mark.xfail()
def test_escape_sequence_resulting_in_multiple_keypresses(parser):
    """Some sequences are interpreted as more than 1 keypress"""
    events = list(parser.feed("\x1b[2;4~"))
    assert len(events) == 2
    assert events[0].key == "escape"
    assert events[1].key == "shift+insert"


@pytest.mark.parametrize("parameter", range(1, 5))
def test_terminal_mode_reporting_synchronized_output_supported(parser, parameter):
    sequence = f"\x1b[?2026;{parameter}$y"
    events = list(parser.feed(sequence))
    assert len(events) == 1
    assert isinstance(events[0], TerminalSupportsSynchronizedOutput)


def test_terminal_mode_reporting_synchronized_output_not_supported(parser):
    sequence = "\x1b[?2026;0$y"
    events = list(parser.feed(sequence))
    assert events == []



from __future__ import annotations

import pytest

from textual.app import App, ComposeResult
from textual.widgets import Label, ListItem, ListView


class EmptyListViewApp(App[None]):
    def compose(self) -> ComposeResult:
        yield ListView()


async def test_listview_pop_empty_raises_index_error():
    app = EmptyListViewApp()
    async with app.run_test() as pilot:
        listview = pilot.app.query_one(ListView)
        with pytest.raises(IndexError) as excinfo:
            listview.pop()
        assert "pop from empty list" in str(excinfo.value)


class ListViewApp(App[None]):
    def __init__(self, initial_index: int | None = None):
        super().__init__()
        self.initial_index = initial_index
        self.highlighted = []

    def compose(self) -> ComposeResult:
        yield ListView(
            ListItem(Label("0")),
            ListItem(Label("1")),
            ListItem(Label("2")),
            ListItem(Label("3")),
            ListItem(Label("4")),
            ListItem(Label("5")),
            ListItem(Label("6")),
            ListItem(Label("7")),
            ListItem(Label("8")),
            initial_index=self.initial_index,
        )

    def _on_list_view_highlighted(self, message: ListView.Highlighted) -> None:
        if message.item is None:
            self.highlighted.append(None)
        else:
            self.highlighted.append(str(message.item.children[0].renderable))


async def test_listview_remove_items() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/4735"""
    app = ListViewApp()
    async with app.run_test() as pilot:
        listview = pilot.app.query_one(ListView)
        assert len(listview) == 9
        await listview.remove_items(range(4, 9))
        assert len(listview) == 4


@pytest.mark.parametrize(
    "initial_index, pop_index, expected_new_index, expected_highlighted",
    [
        (2, 2, 2, ["2", "3"]),  # Remove highlighted item
        (0, 0, 0, ["0", "1"]),  # Remove first item when highlighted
        (8, None, 7, ["8", "7"]),  # Remove last item when highlighted
        (4, 2, 3, ["4", "4"]),  # Remove item before the highlighted index
        (4, -2, 4, ["4"]),  # Remove item after the highlighted index
    ],
)
async def test_listview_pop_updates_index_and_highlighting(
    initial_index, pop_index, expected_new_index, expected_highlighted
) -> None:
    """Regression test for https://github.com/Textualize/textual/issues/5114"""
    app = ListViewApp(initial_index)
    async with app.run_test() as pilot:
        listview = pilot.app.query_one(ListView)

        await listview.pop(pop_index)
        await pilot.pause()

        assert listview.index == expected_new_index
        assert listview._nodes[expected_new_index].highlighted is True
        assert app.highlighted == expected_highlighted


@pytest.mark.parametrize(
    "initial_index, remove_indices, expected_new_index, expected_highlighted",
    [
        (2, [2], 2, ["2", "3"]),  # Remove highlighted item
        (0, [0], 0, ["0", "1"]),  # Remove first item when highlighted
        (8, [-1], 7, ["8", "7"]),  # Remove last item when highlighted
        (4, [2, 1], 2, ["4", "4"]),  # Remove items before the highlighted index
        (4, [-2, 5], 4, ["4"]),  # Remove items after the highlighted index
        (4, range(0, 9), None, ["4", None]),  # Remove all items
    ],
)
async def test_listview_remove_items_updates_index_and_highlighting(
    initial_index, remove_indices, expected_new_index, expected_highlighted
) -> None:
    """Regression test for https://github.com/Textualize/textual/issues/5114"""
    app = ListViewApp(initial_index)
    async with app.run_test() as pilot:
        listview = pilot.app.query_one(ListView)

        await listview.remove_items(remove_indices)
        await pilot.pause()

        assert listview.index == expected_new_index
        if expected_new_index is not None:
            assert listview._nodes[expected_new_index].highlighted is True
        assert app.highlighted == expected_highlighted


if __name__ == "__main__":
    app = ListViewApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Static

TEXT = """I must not fear. Fear is the mind-killer. Fear is the little-death that brings total obliteration. I will face my fear."""


class WrapApp(App):
    CSS_PATH = "text_wrap.tcss"

    def compose(self) -> ComposeResult:
        yield Static(TEXT, id="static1")
        yield Static(TEXT, id="static2")


if __name__ == "__main__":
    app = WrapApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import Static


async def test_compositor_scroll_placements():
    """Regression test for https://github.com/Textualize/textual/issues/5249
    The Static should remain visible.
    """

    class ScrollApp(App):
        CSS = """
        Screen {
            overflow: scroll;
        }
        Container {
            width: 200vw;
        }
        #hello {
            width: 20;
            height: 10;
            offset: 50 10;
            background: blue;
            color: white;
        }
        """

        def compose(self) -> ComposeResult:
            with Container():
                yield Static("Hello", id="hello")

        def on_mount(self) -> None:
            self.query_one("Screen").scroll_to(20, 0, animate=False)

    app = ScrollApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        static = app.query_one("#hello")
        widgets = app.screen._compositor.visible_widgets
        # The static wasn't scrolled out of view, and should be visible
        # This wasn't the case <= v0.86.1
        assert static in widgets
